# -*- coding: utf-8 -*-
"""
Last modified: Nov 22 2024

@author: Hermann Zeyen
University Paris-Saclay

    functions:
       - magnetization_components
       - tandet
       - matrixExtension
       - gradient
       - integral
       - compon
       - mag_color_map
       - data_plot

    Class: Earth_mag with methods:
          - __init__
          - earth_components

"""
import numpy as np
import matplotlib.pyplot as plt
import matplotlib.colors as colors


def magnetization_components(sus, rem, inc, dec, earth):
    """
    Routine calculates the 3 components (tx, ty, tz) of magnetization for
    a combination of induced and remanent magnetization.

    """
    tx = 0.
    ty = 0.
    tz = 0.
    if sus != 0. and earth.f > 0.:
        mag = earth.f*sus
        tx = mag*earth.cdci
        ty = mag*earth.sdci
        tz = mag*earth.sie
# The factor 100 comes from the fact that the calculated anomaly in A/m must be
# multiplied by mu0/4pi = 10**-7 and again by 10**9 to pass from T to nT.
    if rem != 0:
        rem_gauss = rem*100.
        cdr = np.cos(np.radians(dec))
        sdr = np.sin(np.radians(dec))
        cir = np.cos(np.radians(inc))
        sir = np.sin(np.radians(inc))
        ehr = rem_gauss*cir
        tx += ehr*cdr
        ty += ehr*sdr
        tz += rem_gauss*sir
    return tx, ty, tz


def tandet(a, b, t_sum):
    """
    arctan argument addition

    Input:
        a,b : float
                atan arguments to be added
                negative a for subtraction
        t_sum: float
                multiple of pi/2 to be added coming eventually from
                an earlier call

    Returns:
        s: float
            value of summed argument
        t_sum: float
            multiple of pi/2 to be added

    """
    q1 = a+b
    q2 = 1.-a*b
    if np.isclose(q2, 0.):
        if q1 < 0.:
            t_sum -= np.pi/2.
        else:
            t_sum += np.pi/2.
        s = 0.
    else:
        s = q1/q2
        if q2 < 0.:
            if q1 < 0.:
                t_sum -= np.pi
            elif q1 > 0.:
                t_sum += np.pi
    return s, t_sum


def matrixExtension(data):
    """
    Creation of extended matix for 2D Fourier transform.
    The toutine mirrors the lower half of the matrix and adds it at the bottom
    and mirrors the upper half to the top. Equivalently right and left

    Input:
        data : 2D numpy array with data

    Output:
        d : 2D numpy array extended in both directions
        (ny1,nx1): Tuple with starting indices of the original data in matrix d
        (ny2,nx2): Tuple with final indices of the original data in matrix d
        plus one. The original data may thus be retrieved as
        data = d[ny1:ny2,nx1:nx2]

    """
    ny, nx = data.shape
    nx_add = int((nx-2)/2)
    ix_add_right = nx-nx_add-1

    ny_add = int((ny-2)/2)
    d = np.zeros((ny+2*ny_add, nx+2*nx_add))
    iy_add_up = d.shape[0]-ny_add-1

    nx_right = nx_add+nx
    ny_up = ny_add+ny
    edge = np.mean(np.array(list(data[:, 0])+list(data[0, :])
                            + list(data[-1, :])+list(data[:, -1])))
    d[ny_add:ny_up, nx_add:nx_right] = data
    fac = (np.sin(np.arange(nx_add)*np.pi/(2*nx_add)))**2
    fac2 = np.flip(fac)
    j = -1
    for i in range(ny_add, ny_up):
        j += 1
        d[i, :nx_add] = fac*(data[j, 1]-edge)*np.flip(data[j, 1:nx_add+1])+edge
        d[i, nx_right:] = fac2*(data[j, -2]-edge)\
            * np.flip(data[j, ix_add_right:nx-1])+edge
    fac = (np.sin(np.arange(ny_add)*np.pi/(2*ny_add)))**2
    fac2 = np.flip(fac)
    for i in range(d.shape[1]):
        d[:ny_add, i] = fac*(d[1, i]-edge)*np.flip(d[1:ny_add+1, i])+edge
        d[ny_up:, i] = fac2*(d[-2, i]-edge)*np.flip(d[iy_add_up:-1, i])+edge
    return d, (ny_add, nx_add), (ny_add+ny, nx_add+nx)


def gradient(data, dx, dy, direction):
    """
    Routine calculates the derivative of a 2D array using FFT
    For x derivative, coefficients are multiplied with 1j*kx,
    for y derivatives with 1j*ky and for z derivative with np.sqrt(kx**2+ky**2)

    Parameters
    ----------
    data : numpy 2D float array (nr*nc)
        Data to be derivated
    dx : float
        grid step size in x direction.
    dy : float
        grid step size in y direction.
    direction : str, may be "x", "y" or "z"
        Direction in which the dervivative is calculated

    Returns
    -------
    numpy 2D float array with derivatives
        DESCRIPTION.

    """
    d, corner1, corner2 = matrixExtension(data)
    dF = np.fft.fft2(d)
    ny1 = corner1[0]
    nx1 = corner1[1]
    ny2 = corner2[0]
    nx2 = corner2[1]
    ny, nx = d.shape
    kx = np.fft.fftfreq(nx, dx) * 2.*np.pi
    ky = np.fft.fftfreq(ny, dy) * 2.*np.pi
    if direction.lower() == "x":
        s = np.outer(np.ones_like(ky), kx) * 1j
    elif direction.lower() == "y":
        s = np.outer(ky, np.ones_like(kx)) * 1j
    else:
        u = np.outer(np.ones_like(ky), kx)
        v = np.outer(ky, np.ones_like(kx))
        s = np.sqrt(u**2 + v**2)
    dF *= s
    return np.real(np.fft.ifft2(dF)[ny1:ny2, nx1:nx2])


def integral(data, dx, dy, direction):
    """
    Routine calculates the integral of a 2D array using FFT
    For x derivative, coefficients are divided by 1j*kx,
    for y derivatives wby 1j*ky and for z derivative by np.sqrt(kx**2+ky**2)
    Zero dividers (zero wavenumbers) are set to 1.

    Parameters
    ----------
    data : numpy 2D float array (nr*nc)
        Data to be derivated
    dx : float
        grid step size in x direction.
    dy : float
        grid step size in y direction.
    direction : str, may be "x", "y" or "z"
        Direction in which the dervivative is calculated

    Returns
    -------
    numpy 2D float array with derivatives
        DESCRIPTION.

    """
    d, corner1, corner2 = matrixExtension(data)
    dF = np.fft.fft2(d)
    ny1 = corner1[0]
    nx1 = corner1[1]
    ny2 = corner2[0]
    nx2 = corner2[1]
    ny, nx = d.shape
    kx = np.fft.fftfreq(nx, dx) * 2.*np.pi
    ky = np.fft.fftfreq(ny, dy) * 2.*np.pi
    if direction.lower() == "x":
        s = np.outer(np.ones_like(ky), kx) * 1j
    elif direction.lower() == "y":
        s = np.outer(ky, np.ones_like(kx)) * 1j
    else:
        u = np.outer(np.ones_like(ky), kx)
        v = np.outer(ky, np.ones_like(kx))
        s = np.sqrt(u**2 + v**2)
    s[s == 0] = 1
    dF /= s
    return np.real(np.fft.ifft2(dF)[ny1:ny2, nx1:nx2])


def compon(dx, dy, dz, earth):
    """
    Calculate the full horizontal and vertical components of rock field from
    the three directional components

    Parameters
    ----------
    dx : float
        field anomaly in x (NS) direction
    dy : float
        field anomaly in y (EW) direction
    dz : float
        field anomaly in z direction
    earth : class Earth_mag object

    Returns
    -------
    dh : float
        horizontal component of the anomaly
    dt : float
        total field component of the anomaly

    """
# If effect of body is large with respect toe Earth's field, the next three
# lines should be uncommented and the other two lines commented.
    dh = dx*earth.cde + dy*earth.sde
    dt = dx*earth.cdci + dy*earth.sdci + dz*earth.sie
    return dh, dt


def mag_color_map(vmn, vmx, cif=2):
    """
    Set colormap for magnetic data such that white corresponds to 0, but the
    blue and red parts of the colorbar are plotted proportional to the values
    and not, as by default, proportional to the colomap

    Parameters
    ----------
    vmn : float
        Minimum value of data for color scale
    vmx : flag
        Maximum value of data for color scale
    cif : int, optional
        number of ciphers to round to. The default is 2.

    Returns
    -------
    br_map : Matplotlib color scale
        Color scale to be used in magnetic anomaly map plotting
    norm : matplotlib norm corresponding to color scale

    """
    c = cif
# Find number of ciffers equal to or above cif necessary to distinguish
# minimum from maximum rounded value
    while True:
        fac = 10**c
        vmin = np.ceil(vmn*fac)/fac
        vmax = np.floor(vmx*fac)/fac
        if vmax > vmin:
            break
        c += 1
    if np.isclose(vmax, vmin):
        vmin -= 0.5
        vmax += 0.5
    ncol_tot = 1024
    if vmax <= 0.:
        all_colors = plt.cm.Blues_r(np.linspace(0, 1, ncol_tot))
    elif vmin >= 0.:
        all_colors = plt.cm.Reds(np.linspace(0, 1, ncol_tot))
    else:
        ncol_neg = int(-ncol_tot*vmin/(vmax-vmin))
        ncol_pos = ncol_tot-ncol_neg
# Define blue colormap for negative values
        colors_neg = plt.cm.Blues_r(np.linspace(0, 1, ncol_neg))
# Define red colormap for positive values
        colors_pos = plt.cm.Reds(np.linspace(0.05, 1, ncol_pos))
        all_colors = np.vstack((colors_neg, colors_pos))
    br_map = colors.LinearSegmentedColormap.from_list('blue_red', all_colors)
    norm = colors.TwoSlopeNorm(vmin=vmin, vcenter=(vmin+vmax)/2, vmax=vmax)
    return br_map, norm


def data_plot(data, fig, ax, title="", xtitle="Easting [m]",
              ytitle="Northing [m]", cmap="rainbow", norm=None, extent=None,
              cbar_title="", vmin=None, vmax=None, font_ax=12, font_tit=14):
    """
    Plots a raster image of regularly gridded data

    Parameters
    ----------
    data : numpy 2D array of size [ny,nx]
        data to be plotted it is supposed that data[0,0] is in the lower left
        corner (therefore the use of np.flip)
    ax : matplotlib axis
        Axis where to plot the map
    title : str, optional
        Title of the plot. The default is "".
    xtitle : str, optional
        Label of the horizontal axis. The default is "Easting [m]".
    ytitle : str, optional
        Label of the vertical axis. The default is "Northing [m]".
    cmap : matplotlib color map, optional
        Colormap to use. The default is "rainbow".
    norm : matplotlib color norm, optional
        Norm to be used, if given. The default is None.
    extent : list, optional
        limits of the axis labels: [xmin,xmax,ymin,ymax]
        The default is None.
        If None, the limits are set to _0.5 for both directions and to nx-0.5
            for the horizontal axis, ny-0.5 for the vertical axis.
        In order to plot the block limits at the correct position, the lower
            limit ust be 1/2 block width smaller than the minimum coordinate of
            the block centers and the maximum 1/2 block width larger than the
            maximum coordinate of block centers.
    cbar_title : str, optional
        Label of the color bar . The default is "".
    vmin : float, optional
        Minimum value of color bar. The default is None.
        If None, minimum value of data is used
    vmax : float, optional
        Maximum value of color bar. The default is None.
        If None, maximum value of data is used
    font_ax : int, optional. Default: 12
        Font size for axis and color bar annotation
    font_tit : int, optional. Default: 14
        Font size for figure title

    Returns
    -------
    im : matplotlab image
    cbar : matplotlib color bar

    """
    if not extent:
        shape_y, shape_x = data.shape
        extent = [-0.5, shape_x-0.5, 0.5, shape_y-0.5]
    if not vmin:
        vmin = data.min()
    if not vmax:
        vmax = data.max()
    if norm:
        im = ax.imshow(np.flip(data, axis=0), norm=norm, cmap=cmap,
                       aspect='equal', extent=extent, rasterized=True)
    else:
        im = ax.imshow(np.flip(data, axis=0), cmap=cmap, aspect='equal',
                       extent=extent, rasterized=True, vmin=vmin, vmax=vmax)
    ax.set_title(title, fontsize=font_tit)
    ax.set_xlabel(xtitle, fontsize=font_ax)
    ax.set_ylabel(ytitle, fontsize=font_ax)
    ax.tick_params(labelsize=font_ax)
#    cax = ax.inset_axes([0.95, 0.1, 0.05, 0.9])
    cax = ax.inset_axes([1.05, 0.05, 0.05, 0.9], transform=ax.transAxes)
    cbar = fig.colorbar(im, shrink=0.9, cax=cax)
    cbar.set_label(cbar_title, fontsize=font_ax-2)
    cbar.ax.tick_params(labelsize=font_ax-2)
    return im, cbar


def get_extremes(data, width=5):
    """
    Search position of relative minima and maxima on a 1D or2D grid

    Parameters
    ----------
    data : 1D or 2D numpy float array
        Data to be analyzed
    width : int, optional, default: 5
        A maximum is recognized if the value at a point (i,j) is larger than
        or equal to all other values within an area of
        (i-width:i+width, j-width:j+width)

    Returns
    -------
    min_pos : list of tuples
        tuples (i,j) of relative minima positions within array data
        (i is y-direction, j is x-direction.
    max_pos : list of tuples
        Similar to min_pos but for relative maxima.

    """
    max_pos = []
    min_pos = []
    if len(data.shape) == 1:
        nx = len(data)
        for ix in range(width, nx, 2*width):
            d = data[ix-width:ix+width+1]

            pos = np.where(d == np.nanmax(d))
            xmax = ix + pos[0][0] - width
            dmax = data[xmax]
            x1 = max(xmax-width, 0)
            x2 = min(xmax+width, nx)
            dd = data[x1:x2]
            if dd.max() == dmax:
                max_pos.append((xmax,))
            pos = np.where(d == np.nanmin(d))
            xmin = ix + pos[0][0] - width
            x1 = max(xmin-width, 0)
            x2 = min(xmin+width, nx)
            dmin = data[xmin]
            dd = data[x1:x2]
            if dd.min() == dmin:
                min_pos.append((xmin,))
        return min_pos, max_pos
    ny, nx = data.shape
    for ix in range(width, nx, 2*width):
        for iy in range(width, ny, 2*width):
            d = data[iy-width:iy+width+1, ix-width:ix+width+1]
            pos = np.where(d == np.nanmax(d))
            ymax = iy + pos[0][0] - width
            xmax = ix + pos[1][0] - width
            y1 = max(ymax-width, 0)
            y2 = min(ymax+width, ny)
            x1 = max(xmax-width, 0)
            x2 = min(xmax+width, nx)
            dmax = data[ymax, xmax]
            dd = data[y1:y2, x1:x2]
            if dd.max() == dmax:
                max_pos.append((ymax, xmax))

            pos = np.where(d == np.nanmin(d))
            ymin = iy + pos[0][0] - width
            xmin = ix + pos[1][0] - width
            y1 = max(ymin-width, 0)
            y2 = min(ymin+width, ny)
            x1 = max(xmin-width, 0)
            x2 = min(xmin+width, nx)
            dmin = data[ymin, xmin]
            dd = data[y1:y2, x1:x2]
            if dd.min() == dmin:
                min_pos.append((ymin, xmin))
    return min_pos, max_pos



"""
Last modification Jan 16 2025

@author: Hermann Zeyen <hermann.zeyen@universite-paris-saclay.fr>
         Universite Paris-Saclay, France

Wrapper for PyMaGra program
"""

import sys
import os
from pathlib import Path
from PyQt5 import QtWidgets

# import Pymagra as pmg
from Pymagra import PyMaGra

if __name__ == "__main__":
    # dir0 = "./Example"
    dir0 = r"E:\Daten\Fontaines_Salées_BRGM"
    # dir0 = r"D:\Tadjikistan\ZolIZard"

    os.chdir(dir0)

    def my_exception_hook(exctype, value, tracebk):
        """
        Not clear what it is good for.
        The problem is that usually no error messages are passed from
        QT.
        I found this in the internet, but it seems it does not work either...

        Parameters
        ----------
        exctype : TYPE
            DESCRIPTION.
        value : TYPE
            DESCRIPTION.
        tracebk : TYPE
            DESCRIPTION.

        Returns
        -------
        None.

        """
        print(exctype, value, tracebk)
        sys._excepthook(exctype, value, tracebk)
        sys.exit(1)

    if __name__ == "__main__" and __package__ is None:
        file = Path(__file__).resolve()
        parent, top = file.parent, file.parents[0]

        sys.path.append(str(top))
        try:
            sys.path.remove(str(parent))
        # Already removed
        except ValueError:
            pass
        if top not in sys.path:
            sys.path.append(top)

        __package__ = "Pymagra"

    try:
        app = QtWidgets.QApplication(sys.argv)
        sys._excepthook = sys.excepthook
        sys.excepthook = my_exception_hook
        main = PyMaGra.Main(dir0)
        main.w.showMaximized()
        sys.exit(app.exec_())
    except Exception as error:
        print(f"An unexpected exception occurred: {error}.")
        exc_type, exc_obj, exc_tb = sys.exc_info()
        fname = os.path.split(exc_tb.tb_frame.f_code.co_filename)[1]
        print(exc_type, fname, exc_tb.tb_lineno)
        pass



# -*- coding: utf-8 -*-
"""
Last modified on Nov 22, 2024

@author: Hermann Zeyen <hermann.zeyen@gmail.com>
         Université Paris-Saclay

Contains the following functions:
    - plot_geography : Plots geography information is it exists
    - plot_lineaments : Plots lineaments usually measured on tilt angle map
    - histo_plot : Plot data histogram data and allow for interactive choice of
      clipping
    - median_plot : Plots line medians as function of line direction and sensor
    - spectrum_plot : Plots logratihmic power spectrum and linear fits,
      allowing interactive modification of low frequency depths
    - spector1D_plot : Plots depths calculated from spectrum decay of all lines
    - spector2D_plot : Plots map with depths calculated from sliding window
      spectrum decay
    - plot_gradients : Plot first and second vertical as well as horizontal
      gradient maps into three subplots of one figure
    - plot_tilt : Plot tilt angle and tilt angle gradient.
    - ana_histo_plot : Plot cumulated histogram of analytic signal data
    - plot_analytic : Plot map of 2D analytic signal and depth inversion
      results
    - base_plot : Plot diurnal variation data

"""
import numpy as np
from PyQt5 import QtWidgets
import matplotlib.pyplot as plt
from matplotlib.ticker import AutoMinorLocator
from matplotlib import colors
from ..utilities import transforms as trans
from ..utilities import utilities as utils
from ..in_out.dialog import dialog
from .new_window import newWindow


def plot_geography(ax, geography, dfac):
    """
    Plot geography information into axis "ax".

    Parameters
    ----------
    ax : Matplotlib.Axis object
    geography : dictionary with the following entries:
        - "type" : may be "point" or "line"
        - "x" : x coordinate(s) of point or line
        - "y" : y coordinate(s) of point or line
    dfac : Factor by which to multiply coordinates (usually to pass from m to
        km)

    """
    xmin, xmax = ax.get_xlim()
    ymin, ymax = ax.get_ylim()
# Plot a point
    for key in geography.keys():
        x = np.array(geography[key]["x"])*dfac
        y = np.array(geography[key]["y"])*dfac
# If fpoint outside map don't plot
        if geography[key]["type"] == "POINT":
            if x < xmin or x > xmax or y < ymin or y > ymax:
                continue
# Check whether a point has already been plotted. If not, add label to legend
            ax.plot(x, y, "o", color="black")
# Plot a line and check whether a line has already been plotted.
        else:
            ax.plot(x, y, "k")


def plot_lineaments(ax, lineaments, dfac):
    """
    Plot lineaments picked from tilt maps into axis "ax".

    Parameters
    ----------
    ax : Matplotlib.Axis object
    lineaments : dictionary with entries
        - "type" : may be "gravity" or "magnetic"
        - "x" : x coordinate(s) of point or line
        - "y" : y coordinate(s) of point or line
    dfac : float
        Factor by which coordinates are multiplied (usually to pass from m to
            km)
    """
    col = "k"
    for key in lineaments.keys():
        if "gravity" in lineaments[key]["type"]:
            col = "w"
        elif "magnetic" in lineaments[key]["type"]:
            col = "k"
        ax.plot(lineaments[key]["x"]*dfac, lineaments[key]["y"]*dfac,
                col, ls="--", linewidth=3)


def histo_plot(data):
    """
    Plot histogram of data and allow for interactive choice of clipping

    Parameters
    ----------
    data : Instance of class DataContainer
        Data to be treated

    Returns
    -------
    min_fix : float
        Minimum accepted value. None if no value clicked
    max_fix : float
        Maximum accepted value. None if no value clicked
    """
    min_fix = None
    max_fix = None
    histo = newWindow("Clip_histogram", 1500, 1000)
    if data.grad_data:
        ax_histo = histo.fig.subplots(1, 2)
    else:
        ax = histo.fig.subplots(1, 1)
        ax_histo = [ax]
    rmin1 = np.nanquantile(data.sensor1, 0.001)
    rmax1 = np.nanmax(data.sensor1)
    counts1, _, _ = ax_histo[0].hist(data.sensor1, 20, (rmin1, rmax1))
    if data.grad_data:
        ax_histo[0].set_title("Sensor 1")
    else:
        if data.data["type"] == "magnetic":
            ax_histo[0].set_title("Magnetic field")
        else:
            ax_histo[0].set_title("Gravity field")
    if data.grad_data:
        rmin2 = np.nanquantile(data.sensor2, 0.001)
        rmax2 = np.nanmax(data.sensor2)
        ax_histo[1].hist(data.sensor2, 20, (rmin2, rmax2))
        ax_histo[1].set_title("Sensor 2")
    histo.setHelp("click within one of the histograms for minimum value, "
                  + "then for maximum value. Click outside = None")
    histo.show()
# Wait for mouse click to define lower data limit
    while True:
        event = histo.get_event()
        if event.name == "button_press_event":
            break
    if event.xdata:
        if data.grad_data:
            if event.xdata > min(data.sensor1.min(), data.sensor2.min()):
                ax_histo[0].plot([event.xdata, event.xdata],
                                 [0, max(counts1)], "r")
                min_fix = event.xdata
        else:
            if event.xdata > data.sensor1.min():
                ax_histo[0].plot([event.xdata, event.xdata],
                                 [0, max(counts1)], "r")
                min_fix = event.xdata
# Wait for mouse click to define upper data limit
    while True:
        event = histo.get_event()
        if event.name == "button_press_event":
            break
    if event.xdata:
        if data.grad_data:
            if event.xdata < max(data.sensor1.max(), data.sensor2.max()):
                ax_histo[0].plot([event.xdata, event.xdata],
                                 [0, max(counts1)], "r")
                max_fix = event.xdata
        else:
            if event.xdata < data.sensor1.max():
                ax_histo[0].plot([event.xdata, event.xdata],
                                 [0, max(counts1)], "r")
                max_fix = event.xdata
    histo.close()
    return min_fix, max_fix


def median_plot(data, median1_even, median1_odd, median2_even, median2_odd,
                nline_even, nline_odd, txt_even, txt_odd):
    """
    Plot median values of each line as function of line direction and sensor

    Parameters
    ----------
    data : Instance of class DataContainer
        Data to be treated
    median1_even : Numpy 1D float array
        Contains median values for lines in even direction for sensor 1
    median1_odd :  Numpy 1D float array
        Contains median values for lines in odd direction for sensor 1
    median2_even :  Numpy 1D float array
        Contains median values for lines in even direction for sensor 2
    median2_odd :  Numpy 1D float array
        Contains median values for lines in odd direction for sensor 2
    nline_even :  Numpy 1D int array
        Contains line numbers of even lines
    nline_odd : Numpy 1D int array
        Contains line numbers of odd lines
    txt_even : str
        Text for even lines in legend
    txt_odd : str
        Text for odd lines in legend

    Returns
    -------
    None.

    """
    fig_median = newWindow("Medians", 1500, 1000)
    ax_median = fig_median.fig.add_subplot()
    ax_median.plot(
        nline_even, median1_even, color="b", marker="*",
        label=f"sensor1 {txt_even}")
    ax_median.plot(nline_odd, median1_odd, color="c", marker="*",
                   label=f"sensor1 {txt_odd}")
    if median2_even[0] is not None:
        ax_median.plot(
            nline_even, median2_even, color="r", marker="*",
            label=f"sensor2 {txt_even}")
        ax_median.plot(nline_odd, median2_odd, color="orange", marker="*",
                       label=f"sensor2 {txt_odd}")
    ax_median.set_title(f'{data["title"]}: median values of all profiles')
    ax_median.set_xlabel("Number of line")
    if data["type"] == "magnetic":
        ax_median.set_ylabel("Magnetic field [nT]")
    else:
        ax_median.set_ylabel("Gravity field [mGal]")
    ax_median.legend(bbox_to_anchor=(1, 0), loc="lower right")
    fig_median.setHelp("Click ENTER to close")
    fig_median.show()
    while True:
        event = fig_median.get_event()
        if event.name == "key_press_event":
            if event.key == "enter":
                fig_median.close()
                break


def spectrum_plot(amp, k, lpos, depth1, depth2, intercept1, intercept2,
                  max_amps, max_pos, amp2=None, k2=None, lpos2=None,
                  max_amps2=None, max_pos2=None):
    """
    Plot spectrum and linear fits and allow interactive modification of
    low frequency depths

    Parameters
    ----------
    amp : numpy 1D float array
        squared amplitudes of spectrum
    k : numpy 1D float array
        wavenumbers
    lpos : float
        Line position (meters)
    depth1 : float
        Depth determined for low frequencies
    depth2 : float
        Depth determined for high frequencies
    intercept1 : float
        Intercept of regression line for low frequencies
    intercept2 : float
        Intercept of regression line for high frequencies
    half_width : int
        Half with to determine local maxima
    max_amps : numpy 1D float array
        Amplitudes at local maxima
    max_pos : numpy 1D float array
        Wave numbers at local maxima
    amp2 : numpy 1D float array, optional; Default: None
        like amp for a possible plot of second spectrum for comparison.
    k2 : numpy 1D float array, optional; Default: None
        like k for a possible plot of second spectrum for comparison.
    lpos2 : float, optional; Default: None
        like lpos for a possible plot of second spectrum for comparison.
    max_amps2 : numpy 1D float array, optional; Default: None
        like max_amps for a possible plot of second spectrum for comparison.
    max_pos2 : numpy 1D float array, optional; Default: None
        like max_pos for a possible plot of second spectrum for comparison.

    Returns
    -------
    depth1 : float
        Depth determined for low frequencies
    intercept1 : float
        Intercept of regression line for low frequencies

    """
    n_Ny = len(k)
    dk = k[1] - k[0]
    xsplit = (intercept2-intercept1)*0.5/(depth2-depth1)
    isplit = np.argmin(abs(k - xsplit))
    y = np.zeros(n_Ny - 1)
    y[:] = np.nan
    y[:isplit+1] = intercept1 - k[:isplit+1] * depth1*2.
    nmx = min(len(k), len(y))
    y[isplit+2:nmx] = intercept2 - k[isplit+2:nmx] * depth2*2.
    y[isplit + 1] = np.nan
    fig_FFT = newWindow("FFT")
    ax_FFT = fig_FFT.fig.add_subplot()
    ax_FFT.plot(k, amp, "k", label="Spectrum in Y")
    ax_FFT.plot(k[:nmx], y[:nmx], "r", label="Average fit")
    ax_FFT.plot(max_pos, max_amps, "r*")
    xplt1 = np.mean(k[:isplit+1]) + dk
    yplt1 = np.mean(y[:isplit+1])
    xplt2 = np.mean(k[isplit+2:]) + dk
    yplt2 = np.mean(y[isplit+2:])
    ax_FFT.text(xplt1, yplt1, f"{depth1:0.2f}", verticalalignment="bottom",
                horizontalalignment="left")
    ax_FFT.text(xplt2, yplt2, f"{depth2:0.2f}",  verticalalignment="bottom",
                horizontalalignment="left")
    if lpos2 is not None:
        ax_FFT.plot(k2, amp2, "blue", label="Spectrum in X")
        ax_FFT.plot(max_pos2, max_amps2, "g*")
        ax_FFT.set_title(f"Line at {lpos:0.1f}/{lpos2:0.1f}")
        ax_FFT.legend(bbox_to_anchor=(1, 1), loc="upper right",
                      fontsize=10)
    else:
        ax_FFT.set_title(f"Line at {lpos:0.1f}")
    ax_FFT.set_ylabel("ln(Amplitude**2)")
    ax_FFT.set_xlabel("Wavenumber [rad/m]")
    fig_FFT.setHelp("Draw line with left mouse button to get new slope. "
                    + "Single click to close without change")
    fig_FFT.show()
    _, coor_x, coor_y = fig_FFT.follow_line(ax_FFT, release_flag=True)
    if len(coor_x) > 0 and coor_x[0] != coor_x[1]:
        depth1 = (-((coor_y[1] - coor_y[0]) / (coor_x[1] - coor_x[0]))/2.)
        intercept1 = coor_y[0]+depth1*2*coor_x[0]
        ax_FFT.text(coor_x[0], coor_y[0], f"  {depth1:0.2f}",
                    verticalalignment="bottom", horizontalalignment="left")
        fig_FFT.canvas.draw()
        while True:
            event = fig_FFT.get_event()
            if event.name == "button_press_event":
                break
    fig_FFT.close()
    return depth1, intercept1


def spector1D_plot(data, depths1, intercepts1, depths2, intercepts2,
                   direction, half_width):
    """
    Plots Spector 1D results and allows modifying depths interactively

    Parameters
    ----------
    data : Instance of class DataContainer
        Data used for spectral analysis
    depths1 : Numpy 1D float array of length number_of_lines
        Calculated depths for low frequencies
    intercepts1 : Numpy 1D float array of length number_of_lines
        Calculated intercept of linear regression line for low frequencies
    depths2 : Numpy 1D float array of length number_of_lines
        Calculated depths for high frequencies
    intercepts2 : umpy 1D float array of length number_of_lines
        Calculated intercept of linear regression line for high frequencies
    direction : int
        If 0, spectra are calculated in Y direction, else in X direction
    half_width : int
        Half width to determine local maxima positions

    Returns
    -------
    depths1 : Numpy 1D float array of length number_of_lines
        Final depths for low frequencies
    intercepts1 : Numpy 1D float array of length number_of_lines
        Final intercept of linear regression line for low frequencies
    depths2 : Numpy 1D float array of length number_of_lines
        Final depths for high frequencies
    intercepts2 : Numpy 1D float array of length number_of_lines
        Final intercept of linear regression line for high frequencies

    """
    if direction:
        dat = data.sensor1_inter.T
        dsamp = data.x_inter[1] - data.x_inter[0]
        lpos = data.y_inter
    else:
        dat = data.sensor1_inter
        dsamp = data.y_inter[1] - data.y_inter[0]
        lpos = data.x_inter
    n_Ny = int(dat.shape[0]/2)
    while True:
        fig_spector = newWindow("Spector")
        ax_spector = fig_spector.fig.add_subplot()
        ax_spector.plot(lpos, depths1)
        ax_spector.invert_yaxis()
        ax_spector.set_title(
            f'{data.data["title"]}: Average depths from spectral analysis')
        ax_spector.xaxis.set_minor_locator(AutoMinorLocator())
        if direction:
            ax_spector.set_xlabel("Northing of line [m]")
        else:
            ax_spector.set_xlabel("Easting of line [m]")
        ax_spector.set_ylabel("Average depth [m]")
        ax_spector.grid(visible=True, which="both")
        fig_spector.setHelp(
            "Click left mouse button to see spectrum and modify slope;"
            + " right mouse button or ENTER key to finish and close window")
        fig_spector.show()
        while True:
            event = fig_spector.get_event()
            if event.name == "button_press_event":
                break
            elif event.name == "key_press_event":
                if event.key == "enter":
                    break
        if event.name == "key_press_event":
            break
        if event.button != 1:
            break
        if not event.xdata:
            continue
        il = np.argmin(abs(lpos - event.xdata))
        dd, kk = trans.log_spect(data.sensor1_inter[:, il], dsamp, n_Ny)
        max_pos, d, _, _ = utils.min_max(dd, half_width=half_width)
        kkk = kk[max_pos]
        depth1, intercept1 = spectrum_plot(
            dd, kk, lpos[il], depths1[il], depths2[il], intercepts1[il],
            intercepts2[il], d, kkk)
        depths1[il] = depth1
        intercepts1[il] = intercept1
        fig_spector.close()
    fig_spector.close()
    return depths1, intercepts1


def spector2D_plot(data, depths1, intercepts1, depths2, intercepts2, xpos,
                   ypos, half_width, n_Ny, window_len, nwiny, nwinx, color):
    """
    Plots map of sliding window depth estimations from spectrum decay.
    Allows modifying depths interactively.

    Parameters
    ----------
    data : Instance of class DataContainer
        Data used for spectral analysis
    depths1 : Numpy 1D float array
        Calculated depths for low frequencies
    intercepts1 : Numpy 1D float array
        Calculated intercept of linear regression line for low frequencies
    depths2 : Numpy 1D float array
        Calculated depths for high frequencies
    intercepts2 : umpy 1D float array
        Calculated intercept of linear regression line for high frequencies
    xpos, ypos : Numpy 1D float array
        X and Y coordintes of all calculated depths
    half_width : int
        Half width to determine local maxima positions
    n_Ny : int
        Nyquist number used for depths calculations
    window_len : float
        Length of sliding window in meters
    nwiny, nwinx : int
        Number of calculated points in Y and X directions
    color : matplotlib color scale
        Color scale used for map plot

    Returns
    -------
    depths1 : Numpy 1D float array of length number_of_lines
        Final depths for low frequencies
    intercepts1 : Numpy 1D float array of length number_of_lines
        Final intercept of linear regression line for low frequencies

    """
    x = np.copy(data.x_inter)
    y = np.copy(data.y_inter)
    if y.max() > 100000.:
        dfac = 0.001
        x *= 0.001
        y *= 0.001
        dunit = "km"
    else:
        dfac = 1.
        dunit = "m"
    nlin = len(data.lineaments)
    dy = data.y_inter[1]-data.y_inter[0]
    dx = data.x_inter[1]-data.x_inter[0]
    while True:
        fig_spect2 = newWindow("Depths from Spector&Grant")
        ax_spect2 = fig_spect2.plot_images(
            depths1, xpos*dfac, ypos*dfac,
            ptitle=f"Spectral depth solutions\nwindow length: {window_len} m",
            xlabel=f"Easting [{dunit}]", ylabel=f"Northing [{dunit}]",
            clabel="Depth [m]", percent=0.005, c=color)
        if nlin and data.plotLin_flag:
            plot_lineaments(ax_spect2, data.lineaments, dfac)
        if data.geography_flag:
            plot_geography(ax_spect2, data.geography, dfac)
        ax_spect2.set_xlim([np.nanmin(xpos*dfac), np.nanmax(xpos*dfac)])
        ax_spect2.set_ylim([np.nanmin(ypos*dfac), np.nanmax(ypos*dfac)])
        fig_spect2.setHelp(
            "Left click on map: show results of nearest point; "
            + "Right click on map close window and return to main window")
        fig_spect2.show()
        while True:
            event = fig_spect2.get_event()
            if event.name == "button_press_event":
                break
        if not event.xdata:
            continue
        if event.button > 1:
            fig_spect2.close()
            break
        ixclick = np.argmin(abs(xpos*dfac - event.xdata))
        ixpos = np.argmin(abs(data.x_inter - xpos[ixclick]))
        iyclick = np.argmin(abs(ypos*dfac - event.ydata))
        iypos = np.argmin(abs(data.y_inter - ypos[iyclick]))
        data1 = data.sensor1_inter[iypos-nwiny:iypos+nwiny, ixpos]
        dd, kk = trans.log_spect(data1, dy, n_Ny)
        if depths1[iyclick, ixclick]:
            depth1 = depths1[iyclick, ixclick]
            intercept1 = intercepts1[iyclick, ixclick]
            depth2 = depths2[iyclick, ixclick]
            intercept2 = intercepts2[iyclick, ixclick]
        max_pos, d, _, _ = utils.min_max(dd, half_width=half_width)
        kkk = kk[max_pos]
        data2 = data.sensor1_inter[iypos, ixpos-nwinx:ixpos+nwinx]
        dd2, kk2 = trans.log_spect(data2, dx, n_Ny)
        max_pos2, d2, _, _ = utils.min_max(dd2, half_width=half_width)
        kkk2 = kk2[max_pos2]
        depth1, intercept1 = spectrum_plot(
            dd, kk, xpos[ixclick], depth1, depth2, intercept1, intercept2, d,
            kkk, dd2, kk2, ypos[iyclick], d2, kkk2)
        depths1[iyclick, ixclick] = depth1
        intercepts1[iyclick, ixclick] = intercept1
        fig_spect2.close()
    fig_spect2.close()
    return depths1, intercepts1


def plot_gradients(data_c, color="rainbow"):
    """
    Plot first and second vertical as well as horizontal gradient maps into
    three subplots of one figure

    Parameters
    ----------
    data_c : Instance of class DataContainer
        Data containing the gradients
    color : matplotlib color scale, optional. Default: "rainbow"
        Color scale to be used

    Returns
    -------
    None
    """
    x = data_c.x_inter
    y = data_c.y_inter
    if y.max() > 100000.:
        dfac = 0.001
        dunit = "km"
    else:
        dfac = 1.
        dunit = "m"
        nlin = len(data_c.lineaments)
    unit = data_c.unit
    nr, nc = data_c.tilt_ang.shape
    data = data_c.vgrad.reshape(nr, nc, 1)
    data = np.concatenate((data, data_c.vgrad2.reshape(nr, nc, 1)), axis=2)
    data = np.concatenate((data, data_c.hgrad.reshape(nr, nc, 1)), axis=2)
    fig_grad = newWindow("Gradients")
    ax_grad = fig_grad.plot_images(
        data, x*dfac, y*dfac,
        ptitle=["Vertical gradient", "Second vertical derivative",
                "Horizontal gradient"],
        xlabel=[f"Easting [{dunit}]", f"Easting [{dunit}]",
                f"Easting [{dunit}]"],
        ylabel=[f"Northing [{dunit}]", f"Northing [{dunit}]",
                f"Northing [{dunit}]"],
        clabel=[f"Vert. grad. [{unit}/m]",
                f"2nd vert. grad. [{unit}/m2]", f"Hor. grad. [{unit}/m]"],
        c=color, percent=0.005,)
    if nlin and data_c.plotLin_flag:
        plot_lineaments(ax_grad[0], data_c.lineaments, dfac)
        plot_lineaments(ax_grad[1], data_c.lineaments, dfac)
        plot_lineaments(ax_grad[2], data_c.lineaments, dfac)
    if data_c.geography_flag:
        plot_geography(ax_grad[0], data_c.geography, dfac)
        plot_geography(ax_grad[1], data_c.geography, dfac)
        plot_geography(ax_grad[2], data_c.geography, dfac)
    xmin = np.nanmin(x*dfac)
    xmax = np.nanmax(x*dfac)
    ymin = np.nanmin(y*dfac)
    ymax = np.nanmax(y*dfac)
    ax_grad[0].set_xlim([xmin, xmax])
    ax_grad[0].set_ylim([ymin, ymax])
    ax_grad[1].set_xlim([xmin, xmax])
    ax_grad[1].set_ylim([ymin, ymax])
    ax_grad[2].set_xlim([xmin, xmax])
    ax_grad[2].set_ylim([ymin, ymax])
    fig_grad.show()
    return fig_grad


def plot_tilt(data, color="rainbow"):
    """
    Plot tilt angle and tilt angle gradient. Allows tracing interactively
    lineaments.

    The tilt angle gradients are normalized such that only the gradient maxima
    are shown on the map, independent of their actual value.

    Parameters
    ----------
    data : Instance of class DataContainer
        Data containing the gradients
    color : matplotlib color scale, optional. Default: "rainbow"
        Color scale to be used

    Returns
    -------
    None

    """
    if data.y_inter.max() > 100000.:
        dfac = 0.001
        dunit = "km"
    else:
        dfac = 1.
        dunit = "m"
    nlin = len(data.lineaments)
    fig_tilt_ang = newWindow("Tilt angle")
    ax_tilt_ang = fig_tilt_ang.plot_images(
            data.tilt_ang, data.x_inter*dfac, data.y_inter*dfac,
            ptitle=f'{data.title}\nTilt angle',
            xlabel=f"Easting [{dunit}]", ylabel=f"Northing [{dunit}]",
            clabel="Tilt_angle [rad]", c=color)
    if data.geography_flag:
        plot_geography(ax_tilt_ang, data.geography, dfac)
    if nlin and data.plotLin_flag:
        plot_lineaments(ax_tilt_ang, data.lineaments, dfac)
    ax_tilt_ang.set_xlim([np.nanmin(data.x_inter*dfac),
                          np.nanmax(data.x_inter*dfac)])
    ax_tilt_ang.set_ylim([np.nanmin(data.y_inter*dfac),
                          np.nanmax(data.y_inter*dfac)])
    fig_tilt_ang.show()

    while True:
        fig_tilt = newWindow("Tilt angle gradient")
        ax_tilt = fig_tilt.plot_images(
                data.tilt_grd, data.x_inter*dfac, data.y_inter*dfac,
                ptitle=f"{data.title}\nNormalized tilt angle gradient",
                xlabel=f"Easting [{dunit}]", ylabel=f"Northing [{dunit}]",
                clabel="Gradient maxima (n.u.)", c=color)
        if data.geography_flag:
            plot_geography(ax_tilt, data.geography, dfac)
        if nlin and data.plotLin_flag:
            plot_lineaments(ax_tilt, data.lineaments, dfac)
        ax_tilt.set_xlim([np.nanmin(data.x_inter*dfac),
                          np.nanmax(data.x_inter*dfac)])
        ax_tilt.set_ylim([np.nanmin(data.y_inter*dfac),
                          np.nanmax(data.y_inter*dfac)])
        fig_tilt.setHelp(
            "Left click add point, right click finish line (not integrated). "
            + "Immediate right click without chosen point: stop tilt routine. "
            + "Click with mouse wheel eliminates nearest lineament")
        fig_tilt.show()
        event, coor_x, coor_y = fig_tilt.follow_line(ax_tilt)
# If follow_line was exited with the wheel button, this means that no new line
# has been defined. It is then considered that the line nearest to the mouse
# click should be erased
        if event.button == 2:
            if nlin == 0:
                continue
            if not event.xdata or not event.ydata:
                continue
            xc = event.xdata
            yc = event.ydata
            min_lin = list(data.lineaments.keys())[0]
            min_dist = (xc-data.lineaments[min_lin]["x"][0])**2\
                + (yc-data.lineaments[min_lin]["y"][0])**2
            for ll, lin in data.lineaments.items():
                for i, x in enumerate(lin["x"]):
                    dist = (xc-x)**2 + (yc-lin["y"][i])**2
                    if dist < min_dist:
                        min_lin = ll
                        min_dist = dist
            print(f"Erase lineament {min_lin}")
            del data.lineaments[min_lin]
            fig_tilt.close()
            continue
        if len(coor_x) == 0:
            fig_tilt.close()
            fig_tilt_ang.close()
            return
        if len(coor_x) == 1:
            continue
        nlin += 1
        data.lineaments[nlin] = {}
        data.lineaments[nlin]["x"] = np.array(coor_x[:-1])/dfac
        data.lineaments[nlin]["y"] = np.array(coor_y[:-1])/dfac
        data.lineaments[nlin]["type"] = data.data["type"]
        fig_tilt.close()


def ana_histo_plot(edges, cum, title):
    """
    Plot cumulative histogram of analytic signal values of one line.
    Allows choosing interactively a clipping value

    Parameters
    ----------
    edges : numpy 1D float array
        Contains the edges of the histogram subdivisions
    cum : numpy 1D float array
        Contains the values of the cumulated histogram in every subdivision
    title : str
        General title of histogram plot

    Returns
    -------
    None
    """
    fig_q = newWindow("Analytic signal quantiles", 1200, 800)
    ax_q = fig_q.fig.add_subplot()
    ax_q.plot(edges, cum)
    ax_q.set_title(f'{title}: Cumulative sum analytic signal')
    ax_q.set_xlabel("value")
    ax_q.set_ylabel("quantile")
    fig_q.setHelp("Click left mouse button to define threshold")
    fig_q.show()
    while True:
        event = fig_q.get_event()
        if event.name == "button_press_event":
            break
    fig_q.close()
    return event.xdata


def plot_analytic(data, color="rainbow"):
    """
    Plots 2D analytic signal of sensor 1 and inverts the signal for source
    depths.
    The user is asked to define interactively a clipping value below which
    the calculated analytic signal is ignored. For this, one line is chosen
    on the screen for which the cumulated histogram is plotted for clip-value
    definition. For this line, also the data and inversion result are plotted.
    The calculation may be done on lines in X or Y direction.

    Parameters
    ----------
    data : Instance of class DataContainer
        Data containing the gradients
    color : matplotlib color scale, optional. Default: "rainbow"
        Color scale to be used

    Returns
    -------
    None

    """
    xc = np.copy(data.x_inter)
    yc = np.copy(data.y_inter)
    dbar = None
    gna = None
    dmin = None
    # data = data_c.analytic_signal1
    if yc.max() > 100000.:
        dfac = 0.001
        dunit = "km"
        xc *= dfac
        yc *= dfac
    else:
        dfac = 1.
        dunit = "m"
    nlin = len(data.lineaments)
    fig_ana = newWindow("Analytic signal")
    ax_ana = fig_ana.plot_images(
            data.analytic_signal1, xc, yc, ptitle="Analytic signal", c=color,
            xlabel=f"Easting [{dunit}]", ylabel=f"Northing [{dunit}]",
            clabel=f"Analytic signal [{data.unit}/m]", percent=0.005,
            bar_height=0.4)
# If lineaments have been measured and plotting is activated, plot them now
    if nlin and data.plotLin_flag:
        plot_lineaments(ax_ana, data.lineaments, dfac)
    ax_ana.set_xlim([xc.min(), xc.max()])
    ax_ana.set_ylim([yc.min(), yc.max()])
    ax_ana.grid(visible=True, which="both")
    ax_ana.xaxis.set_minor_locator(AutoMinorLocator())
    ax_ana.yaxis.set_minor_locator(AutoMinorLocator())
    fig_ana.setHelp("Click left mouse button to see analytic signal of one "
                    + "line in Y, right: in X; Wheel or press ENTER to finish "
                    + "and close window")
    fig_ana.show()
# Interpretation is based on the squared analytic signal
    d2 = data.analytic_signal1**2
    index = np.isnan(d2)
    d2[index] = 0.0
# Wait for mouse click. Left mouse click will trigger inversion of analytic
#      signal for depths along N-S lines, right mouse click along E-W lines.
#      Clicking on wheel finishes this module.
# The line neares to the mouse click will be shown in an own window with
#     inversion results
    while True:
        while True:
            event = fig_ana.get_event()
            if event.name == "button_press_event":
                break
            if event.name == "key_press_event":
                if event.key == "enter":
                    break
        if event.name == "key_press_event":
            if event.key == "enter":
                fig_ana.close()
                return
        if not event.xdata or not event.ydata:
            continue
        if event.button == 2:
            fig_ana.close()
            return
        if event.button == 1:
            nline = np.argmin(abs(xc-event.xdata))
            pos = yc
            pos_line = xc[nline]
            text = f"{pos_line} East"
            text_x = f"Northing [{dunit}]"
            nl = len(xc)
            direct = "y"
        else:
            nline = np.argmin(abs(yc-event.ydata))
            pos = xc
            pos_line = yc[nline]
            text = f"{pos_line} North"
            text_x = f"Easting [{dunit}]"
            nl = len(yc)
            direct = "x"
        dx = pos[1] - pos[0]
        slope = []
        intercept = []
        depth = []
        alpha = []
        x_center = []
        y_center = []
        fit = []
        pmax = []
# Mask valaues smaller than 1/1000 of maximum values by setting them to
#      negative values
        if dmin is None:
            dmn = d2.max()/1000.
            dmx = d2.max()
            hist, edges = np.histogram(d2.flatten(), bins=1000,
                                       range=(dmn, dmx), density=True)
            dedge = edges[1] - edges[0]
            cum = np.cumsum(hist) * dedge
            nlim = int(len(cum) * 0.85)
            dmin = ana_histo_plot(edges[1:-nlim], cum[:-nlim], data.title)
            depth_min = 0.
            depth_max = dx*20./dfac
        results, okButton = dialog(
            ["Threshold", "Minimum depth", "Maximum depth"],
            ["e", "e", "e"], [dmin, depth_min, depth_max],
            "Analytic signal parameters")
        if okButton:
            dmin = float(results[0])
            depth_min = float(results[1])
            depth_max = float(results[2])
        dd = np.copy(d2)
        dd[d2 < dmin] = -1.0

# prepare window with inversion results of clicked line
        fig_sig = newWindow("Analytic signal inversion", 1200, 800)
# Start loop over lines
        s_best = 1.0e10
        i_best = 0.0
        x_best = 0.0
        dx = pos[1] - pos[0]
        for ll in range(nl):
            if direct == "y":
                d = np.copy(d2[:, ll])
                d_interest = np.copy(d2[:, ll])
                d_interest[d < dmin] = -1.0
                xline = xc[ll]
            else:
                d = np.copy(d2[ll, :])
                d_interest = np.copy(d2[ll, :])
                d_interest[d < dmin] = -1.0
                yline = yc[ll]
# If no data larger than the threshold exist on the line, continue with next
# line
            if d_interest.max() < 0.0:
                continue
# If line corresponds to clicked one, plot analytic signal
            if ll == nline:
                ax_sig = fig_sig.fig.add_subplot()
                ax_sig.plot(pos, d, "k*", label="Analytic signal")
                ax_sig.plot(pos[d_interest > 0], d_interest[d_interest > 0],
                            "g*", label="Unused above threshold")
                ax_sig.set_title(f"Squared analytic signal at {text}")
                ax_sig.set_xlabel(text_x)
                ax_sig.set_ylabel(f"Amplitude [{data.unit}/m]**2")
# # Find maxim and minima along line
            max_pos, _, min_pos, _ = utils.min_max(d_interest,
                                                   half_width=5)
            lab_flag = False
# Start loop over all found maxima
            for p in max_pos:
                ff = np.inf
# Search points belonging to the actual maximum. Limit is found when either
#    a negative value is found (too small analytic signal to be interpreted) or
#    when the next relative minimum is found
                for k in range(p, -1, -1):
                    if d_interest[k] < 0.0 or k in min_pos:
                        break
                n1 = k + 1
                for k in range(p, len(d_interest)):
                    if d_interest[k] < 0.0 or k in min_pos:
                        break
                n2 = k
# If less than 4 points have been found belonging to the maximum, dont threat
#    this maximum
                if n2 - n1 < 3:
                    continue
# Do inversion
# For this the inverse of squared analytic signal should have a linear
#     relation with the squared distance from the maximum.
#     Since the exact position of the maximum is not know, suppose that it is
#     located in the range measured maximum +/- 1/2*data spacing
#     The regression line is calculated using the squared amplitude as weight.
#     In this way, the high amplitudes have a stonger weight for the inversion
#     than the small ones.
#     The obtained slope is the inverse of the parameter alpha, the intercept
#     corresponds to (depth/alpha)**2

# Calculate inverse of squared amplitude (add a small value to avoid
# division by 0)
                y = 1 / (d_interest[n1:n2] + d_interest[p] * 1e-20)
# Test positions around measured maximum with a step of dx/10 for best solution
# x contains the squared distance from maximum
                for x0 in np.arange(pos[p]-dx/2.0, pos[p]+dx/2.0, dx/10):
                    x = ((pos[n1:n2] - x0)/dfac) ** 2
# G is the Frechet matrix
                    G = np.ones((len(x), 2))
                    G[:, 0] = x
# C contains the weights
                    C = np.diag(1.0 / y**2)
# Invert for regression line coefficients
                    mat1 = np.matmul(np.transpose(G), C)
                    coefs = np.matmul(
                        np.matmul(np.linalg.inv(np.matmul(mat1, G)), mat1),
                        y)
                    y_fit = np.matmul(G, coefs)
                    f = np.sum((y - y_fit)**2)
# If fit is better than earlier values, store corresponding parameters, but
#    only if slope and intercept are positive
                    if f < ff and coefs[0] > 0.0 and coefs[1] > 0.0:
                        s_best = coefs[0]
                        i_best = coefs[1]
                        x_best = x0
                        ff = f
# If a best fit was found, store the parameters
                if np.isfinite(ff):
                    dep = np.sqrt(i_best/s_best)-data.h1_sensor
                    if depth_min <= dep <= depth_max:
                        slope.append(s_best)
                        intercept.append(i_best)
                        alpha.append(np.sqrt(1./s_best))
                        depth.append(dep)
                        if direct == "y":
                            y_center.append(x_best)
                            x_center.append(xline)
                        else:
                            y_center.append(yline)
                            x_center.append(x_best)
                        fit.append(np.sqrt(ff))
                        pmax.append(p)
# If line corresponds to picked one, plot results into floating window
                        if ll == nline:
                            xx = ((pos[n1:n2]-x_best)/dfac)**2
                            yy = xx*s_best+i_best
                            yy = 1./yy
                            if lab_flag:
                                ax_sig.plot(pos[n1:n2], d_interest[n1:n2],
                                            "r*")
                                ax_sig.plot(pos[p], d_interest[p], "b*")
                                ax_sig.plot(pos[n1:n2], yy, "r")
                            else:
                                ax_sig.plot(pos[n1:n2], d_interest[n1:n2],
                                            "r*", label="Fitted points")
                                ax_sig.plot(pos[p], d_interest[p], "b*",
                                            label="Fitted maximum")
                                ax_sig.plot(pos[n1:n2], yy, "r",
                                            label="Calculated signal")
                            lab_flag = True
                            ax_sig.text(x_best, yy.max(),
                                        f"h:{depth[-1]:0.1f}, "
                                        + f"al:{alpha[-1]:0.2f}")
            if ll == nline:
                ax_sig.legend(bbox_to_anchor=(1, 1), loc="upper right",
                              fontsize=10)
        slope = np.array(slope)
        intercept = np.array(intercept)
        alpha = np.array(alpha)
        depth = np.array(depth)
        x_center = np.array(x_center)
        y_center = np.array(y_center)
        fit = np.array(fit)
# Store results into file, name depends on the directions used for calculation
        if direct == "y":
            file = "Analytic-signal-solutions_N-S-lines.dat"
        else:
            file = "Analytic-signal-solutions_E-W-lines.dat"
        with open(file, "w", encoding="utf-8") as fo:
            fo.write("     X        Y    Depth[m]   Alpha    Fit\n")
            for i, d in enumerate(depth):
                fo.write(f"{x_center[i]:0.1f} {y_center[i]:0.1f} "
                         + f"{d:0.2f} {alpha[i]:0.3f} {fit[i]:0.6f}\n")
        fig_sig.setHelp("Click any mouse button to close window")
        fig_sig.show()
# Wait for click within results window to close it
        event = fig_sig.get_event()
        fig_sig.close()
# Plot the obtained depths onto map of analytic signal
#            vmin = np.quantile(depth,0.01)
        if dbar is not None:
            dbar.remove()
        if gna is not None:
            gna.remove()
        if len(depth) == 0:
            _ = QtWidgets.QMessageBox.warning(
                None, "Warning", "No depths calculated\n\n"
                + f"Probably you placed the threshold too high ({dmin})\n"
                + "\ntry again with larger threshold.",
                QtWidgets.QMessageBox.Close, QtWidgets.QMessageBox.Close)
            continue
        vmin = depth.min()
        vmax = np.quantile(depth, 0.99)
        nc = -int(np.log10(vmax))+2
        dv = round((vmax-vmin)/10, nc)
        vmin = np.ceil(vmin / dv) * dv
        v = list(np.arange(vmin, vmin+10.5*dv, dv))
        cmap = plt.get_cmap("hot_r")
        norm = colors.BoundaryNorm(v, cmap.N)
        gna = ax_ana.scatter(x_center, y_center, c=depth, cmap=cmap,
                             norm=norm)
# Plot colorbar for depths into the axis
        cax = ax_ana.inset_axes([1.015, 0.6, 0.015, 0.35],
                                transform=ax_ana.transAxes)
        dbar = fig_ana.fig.figure.colorbar(
            gna, cax=cax, extend="max", orientation="vertical",
            anchor=(0, 0), shrink=0.5, aspect=5)
        dbar.ax.tick_params(labelsize=12)
        dbar.ax.set_ylabel("Source depths [m]", fontsize=12)

        ncd = -int(np.log10(dmin))+2
        ax_ana.set_title(
            f"Analytic signal and depth solutions in {direct} "
            + f"direction\nthreshold = {np.round(dmin,ncd)}", fontsize=14)
        fig_ana.setHelp(
            "Click left mouse button to see analytic signal of one "
            + "line in Y, right: in X; Wheel or ENTER to finish and "
            + "close window")
        fig_ana.canvas.draw()


def base_plot(base, data_type="magnetic"):
    """
    Plot base station data
    Allows defining interactively using the right mouse button areas where to
    mute data (data are set to nan)

    Parameters
    ----------
    base : Instance of class Geometrics
        Contains base station data
    data_type : str
        Usually "magnetics", may also be "gravity"

    Returns
    -------
    None
    """
    while True:
        fig_base = newWindow("Diurnal variations", 800, 500)
        ax_base = fig_base.fig.add_subplot()
        ax_base.plot(base.time_base/86400.0, base.base)
        ax_base.set_title("Diurnal variations")
        ax_base.set_xlabel("Time of acquisition [day in year]")
        if data_type == "magnetic":
            ax_base.set_ylabel("Magnetic field [nT]")
        else:
            ax_base.set_ylabel("Gravity field [mGal]")
        fig_base.setHelp("Click left mouse button to close, wheel or right "
                         + "button at two points to cut out data")
        fig_base.show()
        event = fig_base.get_event()
        if event.name == "key_press_event":
            if event.key == "enter":
                break
        else:
            if event.button == 1:
                break
        x0 = event.xdata*86400.0
        while True:
            event = fig_base.get_event()
            if event.name == "button_press_event":
                break
        x1 = event.xdata*86400.0
        index = np.where((base.time_base > min(x0, x1)) &
                         (base.time_base < max(x0, x1)))
        base.base[index] = np.nan
        fig_base.close()
    fig_base.close()
    return base



# -*- coding: utf-8 -*-
# from .plotting import plot, newWindow

# print(f"Invoking __init__.py for {__name__}")

# from . import floating_plots
# from .main_window import mainWindow
from .new_window import newWindow

__all__ = ["newWindow"]



# -*- coding: utf-8 -*-
"""
last modified on Dec 02, 2024

@author: Hermann Zeyen <hermann.zeyen@universite-paris-saclay.fr>
         University Paris-Saclay, France

Needed Python packages:
    PyQt5
    matplotlib
    numpy
    os
    sys
    signal
    sklearn
    scipy

Needed private files:

    Geometrics.py
    Magnetics.ui


Contains the following class:
    Main : Main class, controlling inputs, data treatment and outputs

"""

import os
from copy import deepcopy
from datetime import datetime, date
from signal import signal, SIGINT
import numpy as np
from PyQt5 import QtWidgets
from .in_out import io
from .plotting.main_window import mainWindow
from .plotting import floating_plots as FP
from .utilities import utilities as u
from .in_out.dialog import dialog
from .data.data import DataContainer
from .inversion.inver import inversion
from .inversion.potential_prism import Prism_calc as PP


class Main(QtWidgets.QWidget):
    """
    Main class for PyMaGra Program

    Parameters
    ----------
    dir0 : str (default = None)
        name of working folder

    Methods
    -------
    - __init__ : Initialization
    - readdata : Input data sets
    - check_data : Check if magnetic data contain errors (mainly nulls)
    - correct_time : Correct instrument time
    - file_change : plot another already opened data file
    - join_data : Join all available data sets into one common set.
    - getGeography : Get geography information (towns, geological borders...)
    - oneOfTwo : Extract one line out of two (for lines in alternating
      directions)
    - readBaseData : Input base station data
    - writeStn : Write treated data in Geometrics .stn format
    - writeDat : Write treated data in Geometrics Surfer format
    - writeGXF : Write gridded data in GXF format
    - saveBaseData : Write base station data into Geometrics .str file
    - original : Reset actual data to the original ones
    - plotOriginal : Plot original data set (but keep treated data as actual
      ones)
    - plotActual : Plot actual, treated data to screen
    - plot_gradient : Toggle on/off plotting of gradient data if there are
    - plot_geography : Toggle on/off plotting of geography data
    - plot_lineaments : Toggle on/off plotting of measured lineaments
    - plotLine : Wrapper for plot_Line
    - plot_Line : Plot a lineament onto the data map
    - changeColorScale : Change parameters for Color scale
    - plotBase : Plot base station data
    - plot_median : Plot medians of all lines
    - zooming : (TODO)
    - zoomOut : (TODO)
    - zoomIn : (TODO)
    - zoomInitial : (TODO)
    - save_plot : Save actual plot into .png file
    - diurnal : Estimate diurnal variations from variation of line medians
    - clean : Eliminate outliers from data
    - justify_median : Justify medians of lines measured in opposite directions
    - justify_gauss : Justify gaussian distribution of lines in opposite
      directions
    - interpol : Interpolate data onto a regular grid
    - nan_fill : Interpolate data at positions have nan-value
    - reduce_pole : Reduce magnetic data to the pole
    - log_spect : Calculate lograrithmic spectrum of a data series
    - fit2lines : Fit two slopes to a data series
    - min_max : Search all local minima and maxima in a vector
    - spector_line : Fit lines to lograithmic spectrum and calculate source
      depth
    - spector : Calculate 1D source depths from data spectrum
    - spector2D : Calculate 1D source depths from data spectrum on a 2D grid
    - tilt : Calculate tilt angle of 2D data set
    - continuation : Continue potential field data set upwards
    - analytic_signal : Calculate analytic signal and source depths
    - inver2D : Do 2.5D inversion
    - inver3D : Do 3D inversion
    - synthetic_model : Define synthetic model and calculate its effect
    - Handler : Should handle exeptions - not really working
    - save_lineaments : Saves measured data lineaments into file
    - closeApp : Close application in an ordered way

    """

    def __init__(self, dir0=None):
        super().__init__()

        try:
            os.chdir(dir0)
        except NameError:
            pass
        self.data_read = False
        self.base_files = []
        self.dat = []
        self.dat_ori = []
        self.data_files = []
        self.file_types = []
        self.data_types = []
        self.config_file = []
        self.percent = 0.01
        self.mincol1 = 0.0
        self.maxcol1 = 0.0
        self.mincol2 = 0.0
        self.maxcol2 = 0.0
        self.mincolg = 0.0
        self.maxcolg = 0.0
        self.d_sensor = 0.9
        self.h_sensor = 0.4
        self.height = 1.3
        self.gradient_flag = True
        self.inter_flag = False
        self.config_flag = False
        self.nan_flag = False
        self.geography_flag = False
        self.n_blocks = 0
        self.grid_flag = True
        self.nlineaments = 0
        self.color = "rainbow"
        self.lineaments = {}
        self.treatments = {}
        self.treatments["diurnal"] = False
        self.treatments["clip"] = False
        self.treatments["justify_median"] = False
        self.treatments["justify_Gauss"] = False
        self.treatments["interpol"] = False
        self.treatments["nan_fill"] = False
        self.treatments["pole"] = False
        self.treatments["odd lines"] = False
        self.treatments["even lines"] = False
        self.treatments["up"] = False
        self.line_choice = "all"
        self.dir0 = dir0
        self.fig_base = None
        self.inclination = 62.0
        self.declination = 0.0
        self.field_flag = False
        self.geography = {}
        self.base = None
        self.fig_line = None
        self.ax_line = None
        self.ax_base = None
        self.fig_median = None
        self.ax_median = None
        self.histo = None
        self.ax_histo = None
        self.fig_spector = None
        self.ax_spector = None
        self.fig_FFT = None
        self.ax_FFT = None
        self.fig_spect2 = None
        self.ax_spect2 = None
        self.fig_FFT2 = None
        self.ax_FFT2 = None
        self.fig_grad = None
        self.ax_grad = None
        self.fig_tilt = None
        self.ax_tilt = None
        self.fig_ana = None
        self.ax_ana = None
        self.fig_q = None
        self.ax_q = None
        self.fig_sig = None
        self.ax_sig = None
        self.wait = True
        self.click = False
        self.press = False
        self.event = None
        self.x_event = None
        self.y_event = None
        self.coor_x = 0.0
        self.coor_y = 0.0
        self.line = None
        self.released = False
        self.line_click = None
        self.x_mouse = 0.0
        self.y_mouse = 0.0
        self.mouse = None
        self.xmin = 0.0
        self.ymin = 0.0
        self.xwin = 0.0
        self.ywin = 0.0
        self.day_joint_flag = False
        self.diff_weight = 1.0
        self.dx = 0.0
        self.dy = 0.0
        self.dz = 0.0
        self.sensor1_inter = []
        self.sensor2_inter = []
        self.sensor1_back = []
        self.sensor2_back = []
        self.sensor1_fill = []
        self.sensor2_fill = []
        self.grad_inter = []
        self.grad_fill = []
        self.x_inter = []
        self.y_inter = []
        self.z_inter = []
        self.mask1 = []
        self.mask2 = []
        self.start = [0.0, 0.0]
        self.side = 0
        self.background = None
        self.canvas = None
        self.axl = None
        self.cidmotion = None
        self.cidrelease = None
        self.cidpress = None
        # string_keys gives the number of keys in data that are not numbers
        self.string_keys = 0
        self.w = mainWindow(self)

        # Input data
        self.actual_plotted_file = 0
        self.readData(dir0)
        x = self.dat[0].data[0]["x"]
        y = self.dat[0].data[0]["y"]
        dx = x.max() - x.min()
        dy = y.max() - y.min()
        if dx > dy:
            self.direction = 1
        else:
            self.direction = 0
        if "m" not in self.data_types[0]:
            self.w.save_base.setEnabled(False)
            self.w.readBase.setEnabled(False)
            self.w.basePlot.setEnabled(False)
            self.w.plotGradient.setEnabled(False)
            self.w.diurnalCorrection.setEnabled(False)
            self.w.medianJustify.setEnabled(False)
            self.w.poleReduction.setEnabled(False)

        self.help = QtWidgets.QLabel(self)
        QtWidgets.qApp.installEventFilter(self)

        # Define actions for Menu buttons
        # Actions for menu File
        self.w.addData.triggered.connect(self.readData)
        self.w.saveSTN.triggered.connect(self.writeStn)
        self.w.saveDat.triggered.connect(self.writeDat)
        self.w.saveGXF.triggered.connect(self.writeGXF)
        self.w.readBase.triggered.connect(self.readBaseData)
        self.w.save_base.triggered.connect(self.saveBaseData)
        self.w.Save_plot.triggered.connect(self.save_plot)
        self.w.geography.triggered.connect(self.getGeography)
        self.w.quitAction.triggered.connect(self.close_app)
        # Actions for menu Display
        self.w.originalPlot.triggered.connect(self.plotOriginal)
        self.w.actualPlot.triggered.connect(self.plotActual)
        self.w.change_file.triggered.connect(self.file_change)
        self.w.join.triggered.connect(self.join_data)
        self.w.plotLine.triggered.connect(self.plot_line)
        self.w.basePlot.triggered.connect(self.plotBase)
        self.w.medianPlot.triggered.connect(self.plot_median)
        self.w.plotGradient.triggered.connect(self.plot_gradient)
        self.w.plotGeo.triggered.connect(self.plot_geography)
        self.w.plotGrid.triggered.connect(self.plot_grid)
        self.w.plotLineaments.triggered.connect(self.plot_lineaments)
        self.w.fill.triggered.connect(self.nan_fill)
        # self.w.zoom.triggered.connect(self.zooming)
        # self.w.zoom_Out.triggered.connect(self.zoomOut)
        # self.w.zoom_In.triggered.connect(self.zoomIn)
        # self.w.zoom_Initial.triggered.connect(self.zoomInitial)
        self.w.changeQuantile.triggered.connect(self.changeColorScale)
        self.w.secondLine.triggered.connect(self.oneOfTwo)
        # Actions for menu Utilities
        self.w.originalData.triggered.connect(self.original)
        self.w.cleanData.triggered.connect(self.clean)
        self.w.timeCorrect.triggered.connect(self.correct_time)
        # self.w.adjust.triggered.connect(self.block_adjust)
        self.w.diurnalCorrection.triggered.connect(self.diurnal)
        self.w.medianJustify.triggered.connect(self.justify_median)
        self.w.gaussJustify.triggered.connect(self.justify_gauss)
        self.w.interpolate.triggered.connect(self.interpol)
        self.w.poleReduction.triggered.connect(self.reduce_pole)
        self.w.lineFFT.triggered.connect(self.spector)
        self.w.Spector_2D.triggered.connect(self.spector2D)
        self.w.tiltAngle.triggered.connect(self.tilt)
        self.w.prolongation.triggered.connect(self.continuation)
        self.w.analytic.triggered.connect(self.analytic_signal)
        # Actions for menu Inversion
        self.w.inv2D.triggered.connect(self.inver2D)
        self.w.inv3D.triggered.connect(self.inver3D)
        self.w.synthetic.triggered.connect(self.synthetic_model)

        # Check whether file "lineaments.dat" exists containing magnetic or
        # gravity lineaments picked from tilt angle maps. If so, create
        # dictionary xith all lineament information.
        if os.path.isfile("lineaments.dat"):
            self.lineaments = io.read_lineaments("lineaments.dat")
            self.nlineaments = len(self.lineaments)
            self.w.plotLineaments.setEnabled(True)
            self.w.plotLin_flag = True
            self.w.lineaments = self.lineaments
            for d in self.dat:
                d.lineaments = self.lineaments
                d.lineament_flag = True
                d.plotLin_flag = True
            self.data.lineaments = self.lineaments
            self.data.lineament_flag = True
            self.data.plotLin_flag = True
        # Intercept CTL-C to exit in a controlled way
        signal(SIGINT, self.Handler)
        self.w.grad_flag = False
        self.grad_flag = False
        self.gradient_flag = False
        if self.data_types[self.actual_plotted_file] == "magnetic":
            self.unit = "nT"
        else:
            self.unit = "mGal"
        # Plot data
        self.fig, self.ax = self.w.plot_triang(
            self.data,
            title=f"{self.data.data['title']}",
            percent=self.percent,
            mincol1=self.mincol1,
            maxcol1=self.maxcol1,
            mincol2=self.mincol2,
            maxcol2=self.maxcol2,
            mincolg=self.mincolg,
            maxcolg=self.maxcolg,
            grad_flag=self.gradient_flag)

    def readData(self, dir0):
        """
        Reads additional data files to the ones read during init of Geometrics
        normally, Geometrics stn files should be read, but another format is
        also possible, considered to be output of program mgwin.
        See Geometrics.read_txt for more information

        Returns
        -------
        None.

        """
        df, ft, dt, self.dir0 = io.get_files(dir0)
        self.data_files += df
        self.file_types += ft
        self.data_types += dt
        ld = len(self.dat)
        # Define the first read data file as active one
        self.actual_plotted_file = ld
        for i, f in enumerate(df):
            fconfig = os.path.basename(f)
            j = fconfig.rfind(".")
            # Check whether there is a configuration file for each data file
            if j > 0:
                fconfig = fconfig[:j] + ".config"
            else:
                fconfig += ".config"
            self.config_file.append(fconfig)
            tcorr_flag = False
            # If there is, read its content
            if os.path.isfile(fconfig):
                self.config_flag = True
                with open(fconfig, "r") as fi:
                    lines = fi.readlines()
                self.title_text = lines[1][:-1]
                self.line_dec = float(lines[2])
                self.height_sens1 = float(lines[3])
                self.height_sens2 = float(lines[4])
                self.sensor_disposition = int(lines[5])
                self.strength = float(lines[6])
                self.inclination = float(lines[7])
                self.declination = float(lines[8])
                if len(lines) > 9:
                    tcorr_flag = True
                    tcorr = float(lines[9])
            else:
                self.config_flag = False
            self.n_blocks += 1

            # Create a data class for the actual file
            self.dat.append(DataContainer(self.n_blocks))
            self.dat[-1].set_values(file=df[i], ftype=ft[i], dtype=dt[i])

            # Read Geometics (*.stn) file
            if ft[i] == "GEOMETRICS":
                if len(self.dat) > 1:
                    self.dat[-1].set_values(
                        line_declination=self.dat[-2].line_declination,
                        h_sensor=self.dat[-2].h1_sensor,
                        h2_sensor=self.dat[-2].h2_sensor,
                        d_sensor=self.dat[-2].d_sensor,
                        dispo=self.dat[-2].dispo)
                if self.config_flag:
                    self.dat[-1].read_geometrics(
                        f, self.height_sens1, self.height_sens2,
                        self.sensor_disposition, self.line_dec,
                        self.title_text)
                else:
                    self.dat[-1].read_geometrics(f)
                self.w.saveSTN.setEnabled(True)
                self.w.fill.setEnabled(True)
            # Read GXF format file
            elif ft[i] == "GXF":
                if self.config_flag:
                    self.dat[-1].read_gxf(
                        f, self.height_sens1, self.line_dec, self.title_text)
                else:
                    self.dat[-1].read_gxf(f)
                self.gradient_flag = False
                self.w.fill.setEnabled(True)
                self.w.plotGradient.setEnabled(False)
# Read BRGM aeromagnetic data
            elif ft[i] == "BRGM":
                self.dat[-1].read_BRGM_flight(f)
                self.gradient_flag = False
                self.w.fill.setEnabled(True)
                self.w.plotGradient.setEnabled(False)
                self.w.poleReduction.setEnabled(False)
# Read *.dat data (usually coming from program MGWIN)
            elif ft[i] == "MGWIN":
                if self.config_flag:
                    self.dat[-1].read_txt(
                        f, self.height_sens1, self.height_sens2,
                        self.line_dec, self.title_text)
                else:
                    self.dat[-1].read_txt(f)
                if self.dat[-1]["grad_data"]:
                    self.gradient_flag = True
                    self.w.plotGradient.setEnabled(True)
                else:
                    self.gradient_flag = False
                    self.w.plotGradient.setEnabled(False)
                self.w.fill.setEnabled(True)
            self.dat[-1].set_values(
                block=self.n_blocks, block_name=f"{os.path.basename(df[i])}",
                typ=self.data_types[i], inter_flag=False,
                treatments=self.treatments, fconfig=fconfig)
            if tcorr_flag:
                self.dat[-1].correct_time(dt=tcorr)
                self.w.timeCorrect.setEnabled(False)
            self.dat_ori.append(deepcopy(self.dat[-1]))
            self.check_data(self.dat[-1].data, f)
# If data are magnetic, no configuration file exits and Earth's field
#   properties were not yet read, ask now for these properties
            line_dir = self.dat[-1].data["line_declination"]
            if "m" in self.data_types[i]:
                if not self.field_flag and not self.config_flag:
                    self.earth = io.get_mag_field(line_dir)
                    self.inclination = self.earth.inc
                    self.declination = self.earth.dec
                    self.strength = self.earth.f
                    self.field_flag = True
# If a configuration file exists but Earth's field properties were not yet
#   defined, create earthMag calss and store the field data.
                elif not self.field_flag:
                    self.earth = io.get_mag_field(
                        line_dir, self.strength, self.inclination,
                        self.declination)
                    self.field_flag = True
            else:
                self.earth = io.get_mag_field(line_dir, 0.0, 0.0, 0.0)
            self.inclination_ori = self.inclination
            self.declination_ori = self.declination
            self.earth_ori = deepcopy(self.earth)
            self.dat[-1].set_values(earth=self.earth)
# If the configuration file did not yet exist, create it now with the data
#   read in via the dialog boxes
            if not self.config_flag:
                with open(fconfig, "w") as fo:
                    fo.write(f"{self.data_types[i]}\n")
                    fo.write(f'{self.dat[-1].data["title"]}\n')
                    fo.write(f'{self.dat[-1].data["line_declination"]}\n')
                    fo.write(f'{self.dat[-1].data["height"]}\n')
                    if self.dat[-1].data["grad_data"]:
                        self.height2 = self.dat[-1].data["height2"]
                        self.dist_sensors = self.dat[-1].data["height"] -\
                            self.height2
                    else:
                        self.height2 = self.dat[-1].data["height"]
                        self.dist_sensors = 0.0
                    fo.write(f"{self.height2}\n")
                    fo.write(f'{self.dat[-1].data["dispo"]}\n')
                    fo.write(f"{self.earth.f}\n")
                    fo.write(f"{self.earth.inc}\n")
                    dec = self.earth.dec
                    fo.write(f'{dec+self.dat[-1].data["line_declination"]}\n')
        if len(self.dat) > 1:
            self.w.change_file.setEnabled(True)
            self.w.join.setEnabled(True)
        self.data = deepcopy(self.dat[self.actual_plotted_file])
        self.height = self.data.data["height"]
        self.string_keys = 0
        data_keys = list(self.data.data.keys())
        for i, k in enumerate(data_keys):
            if isinstance(k, (str)):
                break
        self.string_keys = len(data_keys) - i
# Plot data from active file
        if ld > 0:
            self.plotActual()

    def check_data(self, data, file):
        """
        Check magnetic data for consitency

        Data acquired with PPM or Cs/Rb magnetometers may have null measurement
        if the sensors are placed parallel or perpendicular to the Earth's
        field. Program checks whether the minimum or maximum values in a data
        set are smaller/larger than the median by more than 20000 nT and if so,
        it gives a warning message proposing to clean up the data using
        Utilities -> Clean data before any other treatment.

        Parameters
        ----------
        data : dictionary
            Contains all data of a data set. entries "s1", "s2" and "grad_data"
            are used
        file : str
            File name of the data set (only used in the warning message).

        Returns
        -------
        None.

        """
        vmin1 = 1000000.0
        vmax1 = -1000000.0
        vmin2 = 1000000.0
        vmax2 = -1000000.0
        vmed = data[0]["median1"]
        err_flag = False
        grad_test = data["grad_data"]
        for key, val in data.items():
            if isinstance(key, (str)):
                break
            vmin1 = min(val["s1"].min(), vmin1)
            vmax1 = max(val["s1"].max(), vmax1)
            if grad_test:
                vmin2 = min(val["s2"].min(), vmin2)
                vmax2 = max(val["s2"].max(), vmax2)
        if vmax1 > vmed + 20000.0 or vmin1 < vmed - 20000.0:
            err_flag = True
        if grad_test:
            if vmax2 > vmed + 20000.0 or vmin2 < vmed - 20000.0:
                err_flag = True
        if err_flag:
            if grad_test:
                _ = QtWidgets.QMessageBox.warning(
                    None, "Warning", f"File {file}:\n\n"
                    + f"Sensor 1: min: {vmin1:0.2f}, max: {vmax1:0.2f}"
                    + f"\nSensor 2: min: {vmin2:0.2f}, max: {vmax2:0.2f}"
                    + "\n\nConsider cleaning up data as first step.",
                    QtWidgets.QMessageBox.Close, QtWidgets.QMessageBox.Close)
            else:
                _ = QtWidgets.QMessageBox.warning(
                    None, "Warning", f"File {file}:\n\n"
                    + f"Data: min: {vmin1:0.2f}, max: {vmax1:0.2f}"
                    + "\n\nConsider cleaning up data as first step.",
                    QtWidgets.QMessageBox.Close, QtWidgets.QMessageBox.Close)

    def correct_time(self, dt=None):
        """
        Wrapper for measurement time correction

        Parameters
        ----------
        dt : float, optional, Default: None
            Time in seconds to be added to the actual time. If None, dt is
            asked for interactively.

        Returns
        -------
        None.

        """
        self.data.correct_time(dt)

    def file_change(self):
        """
        Choose another file to be represented and treated.

        Dialogue box presents a list of available files other than the one
        actually on the screen and allows to choose by clicking a radio button.
        Clicking on Cancel keeps the actual data.

        Returns
        -------
        None.

        """
        self.dat[self.actual_plotted_file] = deepcopy(self.data)
        labels = [[]]
        file_nr = []
        for i, d in enumerate(self.dat):
            if i == self.actual_plotted_file:
                continue
            labels[0].append(f"{d.data['block_name']}")
            file_nr.append(i)
        results, ok_button = dialog(labels, "r", "0", title="choose data file")
        if not ok_button:
            return None
        self.actual_plotted_file = file_nr[int(results[0])]
        self.data = deepcopy(self.dat[self.actual_plotted_file])
        self.inter_flag = self.data.inter_flag
        if self.inter_flag:
            self.x_inter = np.copy(self.data.x_inter)
            self.y_inter = np.copy(self.data.y_inter)
            self.sensor1_inter = np.copy(self.data.sensor1_inter)
            if self.data.data["grad_data"]:
                self.sensor2_inter = np.copy(self.data.sensor2_inter)
                self.grad_inter = np.copy(self.data.grad_inter)
        self.treatments = deepcopy(self.data.treatments)
        self.plotActual()

    def join_data(self):
        """
        Joins all available data sets into one common data set

        All data must have the same data type (magnetic vs gravity) and be
        located in the same coordinate system. If the different data blocks are
        not contiguous in space, program will interpolate meaningless "data".

        The original data sets are maintained such that using file_change, one
        may return to one of the smaller data sets. If later a new data set is
        added and data are joint again, the joint data set is excluded.

        Returns
        -------
        None.

        """
        self.dat[self.actual_plotted_file] = deepcopy(self.data)
        nbk = len(self.dat)
        self.n_blocks += 1
        self.dat.append(DataContainer(self.n_blocks))
        nlines = 0
        blkn = "blocks"
        self.dat[-1].data = deepcopy(self.dat[0].data)
        for key, val in self.dat[0].data.items():
            if isinstance(key, str):
                del self.dat[-1].data[key]
        pos_lines = []
        blkn = f"{blkn} {self.dat[0].data['block']}"
        for key, val in self.dat[0].data.items():
            if isinstance(key, (str)):
                break
            if val["direction"] in ("N", "S", 0.0, 180.0):
                pos_lines.append(np.median(val["x"]))
            else:
                pos_lines.append(np.median(val["y"]))
        nlines = len(pos_lines)
        self.treatments = deepcopy(self.dat[0].treatments)
        for i, d in enumerate(self.dat[1:-1]):
            if d.treatments != self.treatments:
                _ = QtWidgets.QMessageBox.warning(
                    None, "Warning", f"Data set {d.data['block']}:\n\n"
                    + "Data treatments are not the same as in other blocks\n\n"
                    + "All blocks must have been treated in the same way "
                    + "before joining\n\nJoin data sets is aborted",
                    QtWidgets.QMessageBox.Close, QtWidgets.QMessageBox.Close)
                self.n_blocks -= 1
                del self.dat[-1]
                return
            if i == nbk:
                break
# Skip already joint data sets (they contain the character "+" in their name)
            if "+" in d.data["block_name"]:
                continue
            blkn = f"{blkn}+{d.data['block']}"
            self.dat[-1].xmin = min(d.xmin, self.dat[-1].xmin)
            self.dat[-1].xmax = max(d.xmax, self.dat[-1].xmax)
            self.dat[-1].ymin = min(d.ymin, self.dat[-1].ymin)
            self.dat[-1].ymax = max(d.ymax, self.dat[-1].ymax)
            for key, val in d.data.items():
                add_flag = False
                if isinstance(key, (str)):
                    break
                if val["direction"] in ("N", "S", 0.0, 180.0):
                    pos_line = np.median(val["x"])
                else:
                    pos_line = np.median(val["y"])
                for il, p in enumerate(pos_lines):
                    if np.isclose(pos_line, p):
                        add_flag = True
                        break
                if add_flag:
                    self.dat[-1].data[il]["x"] = np.concatenate(
                        (self.dat[-1].data[il]["x"], val["x"]))
                    self.dat[-1].data[il]["y"] = np.concatenate(
                        (self.dat[-1].data[il]["y"], val["y"]))
                    if val["direction"] in ("N", "S", 0.0, 180.0):
                        index = np.argsort(self.dat[-1].data[il]["y"])
                    else:
                        index = np.argsort(self.dat[-1].data[il]["x"])
                    self.dat[-1].data[il]["x"] =\
                        self.dat[-1].data[il]["x"][index]
                    self.dat[-1].data[il]["y"] =\
                        self.dat[-1].data[il]["y"][index]
                    self.dat[-1].data[il]["z"] = np.concatenate(
                        (self.dat[-1].data[il]["z"], val["z"]))[index]
                    self.dat[-1].data[il]["s1"] = np.concatenate(
                        (self.dat[-1].data[il]["s1"], val["s1"]))[index]
                    self.dat[-1].data[il]["s2"] = np.concatenate(
                        (self.dat[-1].data[il]["s2"], val["s2"]))[index]
                    self.dat[-1].data[il]["time"] = np.concatenate(
                        (self.dat[-1].data[il]["time"], val["time"]))[index]
                    if self.dat[-1].data[il]["direction"] in ("N", "S", 0.0,
                                                              180.0):
                        _, index = np.unique(
                            self.dat[-1].data[il]["y"], return_index=True)
                    else:
                        _, index = np.unique(
                            self.dat[-1].data[il]["x"], return_index=True)
                    self.dat[-1].data[il]["x"] =\
                        self.dat[-1].data[il]["x"][index]
                    self.dat[-1].data[il]["y"] =\
                        self.dat[-1].data[il]["y"][index]
                    self.dat[-1].data[il]["z"] =\
                        self.dat[-1].data[il]["z"][index]
                    self.dat[-1].data[il]["s1"] =\
                        self.dat[-1].data[il]["s1"][index]
                    self.dat[-1].data[il]["s2"] =\
                        self.dat[-1].data[il]["s2"][index]
                    self.dat[-1].data[il]["time"] =\
                        self.dat[-1].data[il]["time"][index]
                else:
                    self.dat[-1].data[nlines] = deepcopy(val)
                    nlines += 1
                    pos_lines.append(pos_line)

            self.dat[-1].sensor1 = np.concatenate((self.dat[-1].sensor1,
                                                   d.sensor1))
            self.dat[-1].sensor2 = np.concatenate((self.dat[-1].sensor2,
                                                   d.sensor2))
            self.dat[-1].x = np.concatenate((self.dat[-1].x, d.x))
            self.dat[-1].y = np.concatenate((self.dat[-1].y, d.y))
            self.dat[-1].z = np.concatenate((self.dat[-1].z, d.z))
            self.dat[-1].topo = np.concatenate((self.dat[-1].topo, d.topo))
            self.dat[-1].time = np.concatenate((self.dat[-1].time, d.time))
        for key, val in self.dat[0].data.items():
            if isinstance(key, (str)):
                self.dat[-1].data[key] = deepcopy(self.dat[0].data[key])
        self.dat[-1].grad_data = self.dat[0].data["grad_data"]
        self.dat[-1].data["block"] = self.n_blocks
        self.dat[-1].data["block_name"] = blkn
        self.dat[-1].data["title"] += f", {blkn}"
        self.dat[-1].data["type"] = self.data_types[0]
        self.dat[-1].line_declination = self.dat[0].line_declination
        self.dat[-1].h_sensor = self.dat[0].h1_sensor
        self.dat[-1].d_sensor = self.dat[0].d_sensor
        self.dat[-1].dispo = self.dat[0].dispo
        self.dat[-1].data["grad_data"] = self.dat[0].data["grad_data"]
        self.dat[-1].data["year"] = self.dat[0].data["year"]
        self.dat[-1].data["height"] = self.dat[0].data["height"]
        self.dat[-1].data["line_declination"] =\
            self.dat[0].data["line_declination"]
        if self.dat[-1].data["grad_data"]:
            self.dat[-1].data["height2"] = self.dat[0].data["height2"]
            self.dat[-1].data["d_sensor"] = self.dat[0].data["d_sensor"]
            self.dat[-1].inter_flag = False
        if self.dat[0].inter_flag:
            self.join_gridded()
        self.dat[-1].treatments = deepcopy(self.treatments)
        self.dat_ori.append(deepcopy(self.dat[-1]))
        self.data_types.append(self.data_types[0])
        self.file_types.append(self.file_types[0])
        self.data_files.append(self.dat[-1].data["block_name"])
        self.data = deepcopy(self.dat[-1])
        self.data.prepare_gdata(original_fill=True)
        self.w.join.setEnabled(False)
        self.w.adjust.setEnabled(True)
        self.actual_plotted_file = len(self.dat) - 1
        self.plotActual()

    def join_gridded(self):
        self.dat[self.actual_plotted_file] = deepcopy(self.data)
        xmin = self.dat[0].x_inter.min()
        xmax = self.dat[0].x_inter.max()
        ymin = self.dat[0].y_inter.min()
        ymax = self.dat[0].y_inter.max()
        self.dx = self.dat[0].x_inter[1] - self.dat[0].x_inter[0]
        self.dy = self.dat[0].y_inter[1] - self.dat[0].y_inter[0]
        for i, d in enumerate(self.dat[1:-1]):
            if "+" in d.data["block_name"]:
                continue
            if not d.inter_flag:
                _ = QtWidgets.QMessageBox.warning(
                    None,
                    "Warning",
                    f"Data set {d.data['block']} has not been gridded"
                    + "Interpolate the joint data set but be aware that\nif "
                    + "there are areas without data,\nthey may be interpolated"
                    + "with meaningless data.",
                    QtWidgets.QMessageBox.Close,
                    QtWidgets.QMessageBox.Close,
                )
                return
            xmin = min(xmin, d.x_inter.min())
            xmax = max(xmax, d.x_inter.max())
            ymin = min(ymin, d.y_inter.min())
            ymax = max(ymax, d.y_inter.max())
        self.x_inter = np.arange(xmin, xmax + self.dx / 2.0, self.dx)
        self.y_inter = np.arange(ymin, ymax + self.dy / 2.0, self.dy)
        nx = len(self.x_inter)
        ny = len(self.y_inter)
        self.sensor1_inter = np.zeros((ny, nx))
        self.sensor1_inter[:, :] = np.nan
        if self.dat[-1].data["grad_data"]:
            self.sensor2_inter = np.zeros((ny, nx))
            self.sensor2_inter[:, :] = np.nan
        for i, d in enumerate(self.dat[:-1]):
            dx = d.x_inter[1] - d.x_inter[0]
            dy = d.y_inter[1] - d.y_inter[0]
            if not (np.isclose(dx, self.dx) and np.isclose(dy, self.dy)):
                _ = QtWidgets.QMessageBox.warning(
                    None,
                    "Warning",
                    f"Data set {d.data['block']}:\n\nGrid step is not like "
                    + f"others:\ndx = {dx}; required: {self.dx}\n"
                    + f"dy = {dy}; required: {self.dy}\n\n"
                    + "Interpolate all blocks with the same grid step and try"
                    + "again.\n\nJoin data sets is aborted",
                    QtWidgets.QMessageBox.Close,
                    QtWidgets.QMessageBox.Close,
                )
                self.n_blocks -= 1
                del self.dat[-1]
                return
            nx1 = int((d.x_inter.min() - xmin) / self.dx)
            nx2 = nx1 + len(d.x_inter)
            ny1 = int((d.y_inter.min() - ymin) / self.dy)
            ny2 = ny1 + len(d.y_inter)
            self.sensor1_inter[ny1:ny2, nx1:nx2] = np.copy(d.sensor1_inter)
            if self.dat[-1].data["grad_data"]:
                self.sensor2_inter[ny1:ny2, nx1:nx2] = np.copy(d.sensor2_inter)
        self.dat[-1].x_inter = np.copy(self.x_inter)
        self.dat[-1].y_inter = np.copy(self.y_inter)
        self.dat[-1].sensor1_inter = np.copy(self.sensor1_inter)
        if self.dat[-1].data["grad_data"]:
            self.grad_inter = (self.sensor2_inter - self.sensor1_inter) /\
                self.dat[-1].data["d_sensor"]
            self.dat[-1].sensor2_inter = np.copy(self.sensor2_inter)
            self.dat[-1].grad_inter = np.copy(self.grad_inter)
        self.dat[-1].inter_flag = True
        self.inter_flag = True
        self.data = deepcopy(self.dat[-1])
        self.data.prepare_gdata(original_fill=True)

    def getGeography(self):
        """
        Ask for file containing geography data, read data and plot them onto
        data map.

        Geography file has the following form:

        #Keyword (may be "#LINE", "#POINT" or "#END")
        If "#POINT", one line follows with x y text \
        If "#LINE", one line follows for every point defining the line x y

        The file must be finished with a line containing #END in the first
        four columns.
        Usually, points are towns and text the name of the town
        x and y must be given in the same coordinate system as the data.

        Returns
        -------
        None.

        """
        files = list(
            QtWidgets.QFileDialog.getOpenFileNames(
                None,
                "Select geography data file",
                "",
                filter="txt (*.txt) ;; all (*.*)",
            )
        )
        if len(files) == 0:
            print("No file chosen, try again")
        if len(files[0]) < 1:
            print("getGeograpgy: No files read")
            return
        self.geography = io.read_geography_file(files[0][0])
        for d in self.dat:
            d.set_values(geography=self.geography)
            d.set_values(geography_flag=True)
        self.data.set_values(geography=self.geography)
        self.data.set_values(geography_flag=True)
        self.w.set_geography_flag(True)
        self.geography_flag = True
        self.w.plotGeo.setEnabled(True)
        self.plotActual()

    def oneOfTwo(self):
        """
        Extract every second line starting with line 1 (odd lines) or line 2
        (even lines). May be useful is there is a strong directional effect in
        magnetic data

        Returns
        -------
        None.

        """
        if self.treatments["odd lines"]:
            choice = ["even", "all"]
        elif self.treatments["even lines"]:
            choice = ["odd", "all"]
        else:
            choice = ["odd", "even"]
        results, okButton = dialog(
            ["Choose data of", choice], ["l", "r"], [None, 0], "Extract lines"
        )
        if okButton:
            c = int(results[1])
            if self.treatments["odd lines"]:
                self.treatments["odd lines"] = False
                u.extract(self.data, choice[c])
                if c == 0:
                    self.treatments["even lines"] = True
                else:
                    self.treatments["even lines"] = False
            elif self.treatments["even lines"]:
                self.treatments["even lines"] = False
                if c == 0:
                    self.treatments["odd lines"] = True
                else:
                    self.treatments["odd lines"] = False
            else:
                if c == 0:
                    self.treatments["odd lines"] = True
                else:
                    self.treatments["even lines"] = True
            u.extract(self.data, choice[c])
            self.data.treatments = deepcopy(self.treatments)
        else:
            return
        self.plotActual()

    def readBaseData(self):
        """
        Read base station data. These data must be Geometrics PPM files

        Returns
        -------
        None.

        """
        data_files, _, _, _ = io.get_files(ftype="base")
        if len(data_files) > 0:
            for f in data_files:
                self.base_files.append(f)
                self.base = DataContainer(0)
                self.base.read_base(f, self.data.data["year"])
            self.w.basePlot.setEnabled(True)
            self.plotBase()

    def writeStn(self):
        """
        Writes actual (last treated interpolated data) to Geometrics stn file.
        The file name is "data-type"followed by date and hour of file
        production

        Returns
        -------
        None.

        """
        fname = u.file_name(self.data.data["type"], ".stn")
        self.data.write_geometrics(fname)

    def writeDat(self):
        """
        Writes actual (last treated not necessarily interpolated data) to
        Geometrics Surfer (.dat) format file.
        The file name is "data-type"followed by date and hour of file
        production

        Returns
        -------
        None.

        """
        fname = u.file_name(self.data.data["type"], ".dat")
        try:
            isinstance(self.sensor1_inter, np.ndarray)
            if self.data.data["grad_data"]:
                self.data.write_dat(fname)
            else:
                self.daa.write_dat(fname)
        except AttributeError:
            self.data.write_dat(fname)

    def writeGXF(self):
        file = u.file_name(self.data.data["type"], ".gxf")
        x0 = self.x_inter.min()
        y0 = self.y_inter.min()
        dx = self.x_inter[1] - self.x_inter[0]
        dy = self.y_inter[1] - self.y_inter[0]
        io.store_gxf(file, self.sensor1_inter, x0, y0, dx, dy)

    def saveBaseData(self):
        """
        Writes base station data to Geometrics G-856 .stn format file.
        The file name is "base"followed by date and hour of file production

        Returns
        -------
        None.

        """
        fname = u.file_name("base", ".stn")
        self.base.write_base(fname)

    def original(self):
        """
        If data were interpolated, go back to non-interpolated data, all
        earlier treatments are retained. If not, all treatments on data are
        undone.
        If no base station data had been read, but they were produced through
        spline fit of median values, those data are deleted.

        Returns
        -------
        None.

        """
        if not self.data.inter_flag:
            self.data = deepcopy(self.dat_ori[self.actual_plotted_file])
            self.treatments["odd lines"] = False
            self.treatments["even lines"] = False
            if len(self.base_files) == 0:
                try:
                    del self.w.base
                    del self.w.time_base
                except AttributeError:
                    pass
            self.treatments["diurnal"] = False
            self.treatments["clip"] = False
            self.treatments["justify_median"] = False
            self.treatments["justify_Gauss"] = False
            self.treatments["odd lines"] = False
            self.treatments["even lines"] = False
        self.w.originalData.setText("Restart with original data")
        self.w.fill.setEnabled(False)
        self.w.poleReduction.setEnabled(False)
        self.w.tiltAngle.setEnabled(False)
        self.w.lineFFT.setEnabled(False)
        self.w.Spector_2D.setEnabled(False)
        self.w.prolongation.setEnabled(False)
        self.w.analytic.setEnabled(False)
        self.w.gaussJustify.setEnabled(False)
        self.w.saveGXF.setEnabled(False)
        self.treatments["interpol"] = False
        self.treatments["nan_fill"] = False
        self.treatments["pole"] = False
        self.treatments["up"] = False
        self.data.treatments = deepcopy(self.treatments)
        self.inclination = self.inclination_ori
        self.declination = self.declination_ori
        self.earth.inc = self.inclination
        self.earth.dec = self.declination
        self.earth.earth_components()
        self.height = self.h_sensor + self.d_sensor
        self.data.set_values(
            earth=self.earth, treatments=self.treatments, inter_flag=False)
        self.plotActual()

    def plotOriginal(self):
        """
        Plot original data set again

        Returns
        -------
        None.

        """
        data = self.dat_ori[self.actual_plotted_file].data
        self.fig, self.ax = self.w.plot_triang(
            data, title=f"{self.data.data['title']}, ", percent=self.percent,
            mincol1=self.mincol1, maxcol1=self.maxcol1, mincol2=self.mincol2,
            maxcol2=self.maxcol2, mincolg=self.mincolg, maxcolg=self.maxcolg,
            grad_flag=self.gradient_flag, c=self.color)

    def plotActual(self):
        """
        Plot actual data set modified and interpolated or not

        Returns
        -------
        None.

        """
        title = f"{self.data.data['title']};\nTreatments:"
        ntreat = 0
        if self.data.inter_flag:
            for t in self.data.treatments.items():
                if t[1]:
                    if ntreat == 2:
                        title += "\n"
                    title += " " + t[0] + ","
                    ntreat += 1
            if title[-1] == ":":
                title += " None"
            if title[-1] == ",":
                title = title[:-1]

            self.fig, self.ax = self.w.plot_image(
                self.data, title=title, percent=self.percent,
                mincol1=self.mincol1, maxcol1=self.maxcol1,
                mincol2=self.mincol2, maxcol2=self.maxcol2,
                mincolg=self.mincolg, maxcolg=self.maxcolg,
                grad_flag=self.gradient_flag, c=self.color,
                dec=self.data.line_declination)
        else:
            for t in self.data.treatments.items():
                if t[0] in ("interpol", "nan_fill", "pole", "up"):
                    continue
                if t[1]:
                    title += " " + t[0] + ","
            if title[-1] == ":":
                title += " None"
            if title[-1] == ",":
                title = title[:-1]
            self.fig, self.ax = self.w.plot_triang(
                self.data, title=title, percent=self.percent,
                mincol1=self.mincol1, maxcol1=self.maxcol1,
                mincol2=self.mincol2, maxcol2=self.maxcol2,
                mincolg=self.mincolg, maxcolg=self.maxcolg,
                grad_flag=self.gradient_flag, c=self.color)

    def plot_gradient(self):
        """
        Activate or desactivate plot of vertical gradient maps

        Returns
        -------
        None.

        """
        self.gradient_flag = not self.gradient_flag
        if self.gradient_flag:
            self.w.plotGradient.setChecked(True)
        else:
            self.w.plotGradient.setChecked(False)
        self.plotActual()

    def plot_geography(self):
        """
        Activate or deactivate plot of geographical information

        Returns
        -------
        None.

        """
        self.w.toggle_geography_flag()
        for d in self.dat:
            d.set_values(geography_flag=self.w.geography_flag)
        self.data.set_values(geography_flag=self.w.geography_flag)
        self.plotActual()

    def plot_lineaments(self):
        """
        Activate or deactivate plot of measured lineaments

        Returns
        -------
        None.

        """
        self.w.plotLin_flag = not self.w.plotLin_flag
        for d in self.dat:
            d.set_values(plotLin_flag=self.w.plotLin_flag)
        self.data.set_values(plotLin_flag=self.w.plotLin_flag)
        self.plotActual()

    def plot_grid(self):
        """
        Activate or deactivate plot of grid

        Returns
        -------
        None.

        """
        self.w.grid_flag = not self.w.grid_flag
        self.grid_flag = not self.grid_flag
        for d in self.dat:
            d.set_values(grid_flag=self.w.grid_flag)
        self.set_values(grid_flag=self.w.grid_flag)
        self.plotActual()

    def plot_line(self):
        self.plotLine(plot_flag=True)

    def plotLine(self, plot_flag=True):
        """
        Plot one line out of the data set.
        The user is first asked to choose one line by mouse click. This line
        is shown in a floating window.
        Erroneous data may be erased by clicking the mouse wheel at the
        beginning and at the end of the zone to be erased. The program
        recognizes automatically whether first or second sensor has been
        clicked. Left mouse click changes line to the left/below, right mouse
        click goes to the line to the right/above.
        A click outside the two coordinate systems stops the module and deletes
        the flowting window.

        Parameters
        ----------
        plot_flag : bool, optional. Default: True
            If true, data along line are plotted and returned, if False, data
            are returned but nor plotted.

        Returns
        -------
        pos : 1D numpy float array
            contains the coordinates of tha data pointsalong the chosen line
        pos_line : float
            position of the chosen line
        s1 : 1D numpy float array
            Measured data along the line of sensor 1
        s2 : 1D numpy float array
            Measured data along the line of sensor 2. If only one sensor, s2 is
            np.array([None])
        direction : str
            May be "N" or "E", depending on whether the left or right mouse
            button was pressed

        """
        direction = ""
        if self.inter_flag:
            self.w.setHelp(
                "Click left mouse button to choose line in Y "
                + "direction or right button for X-direction"
            )
        else:
            self.w.setHelp("Click mouse button to choose a line")
        # Wait for mouse click to choose line to be plotted first
        while True:
            event = self.w.get_mouse_click(self.fig)
            if event.name == "button_press_event":
                if event.inaxes:
                    break
        pos, pos_line, s1, s2, direction = self.data.plot_line(plot_flag,
                                                               event)
        self.plotActual()
        return pos, pos_line, s1, s2, direction

    def changeColorScale(self):
        """
        Change the limiting quantiles for creating of color scales.

        The values below self.percent and above (1-self.percent) are clipped

        Returns
        -------
        None.

        """
        cols = ["rainbow", "seismic", "viridis", "special mag"]
        if self.w.grad_data:
            results, okButton = dialog(
                [
                    "If min==max, quantile is used\n"
                    + "If quantile is also 0, min and max values of arrays\n"
                    + "If min!=max, these values set the color scale limits",
                    "Cliping quantile for plotting",
                    f"Minimum of color scale sensor1 [{self.unit}]",
                    f"Maximum of color scale sensor1 [{self.unit}]",
                    "_______________________________",
                    f"Minimum of color scale sensor2 [{self.unit}]",
                    f"Maximum of color scale sensor2 [{self.unit}]",
                    "_______________________________",
                    f"Minimum of color scale gradient [{self.unit}/m]",
                    f"Maximum of color scale gradient [{self.unit}/m]",
                    "Color map", cols,],
                ["l", "e", "e", "e", "l", "e", "e", "l", "e", "e", "l", "b"],
                [None, self.percent, self.mincol1, self.maxcol1, None,
                 self.mincol2, self.maxcol2, None, self.mincolg, self.maxcolg,
                 0,], "Color scale limits")
        else:
            results, okButton = dialog(
                ["If min==max, quantile is used\nIf quantile is also 0, min "
                 + "and max values of arrays\nIf min!=max, these values set "
                 + "the color scale limits", "Cliping quantile for plotting",
                 f"Minimum of color scale [{self.unit}]",
                 f"Maximum of color scale [{self.unit}]", "Color map", cols],
                ["l", "e", "e", "e", "l", "b"],
                [None, self.percent, self.mincol1, self.maxcol1, 0],
                "Color scale limits")
        if okButton:
            self.percent = float(results[1])
            self.mincol1 = float(results[2])
            self.maxcol1 = float(results[3])
            if self.w.grad_data:
                self.mincol2 = float(results[5])
                self.maxcol2 = float(results[6])
                self.mincolg = float(results[8])
                self.maxcolg = float(results[9])
                self.color = cols[int(results[11])]
            else:
                self.color = cols[int(results[5])]
            self.plotActual()
        else:
            print(f"\nClipping quantile left at {self.percent:0.3f}\n"
                  + f"    minimum color: {self.mincol1:0.1f} {self.unit}"
                  + f"    maximum color: {self.maxcol1:0.1f} {self.unit}")

    def plotBase(self):
        """
        Plot base station data as function of time (seconds in year).
        The user may erase erroneous data by clicking the mouse wheel at the
        beginning and at the end of the zone to be erased.
        A click with any other mouse button finishes the module and closes the
        floating window.

        Returns
        -------
        None.

        """
        try:
            self.fig_base.clf()
        except (NameError, AttributeError):
            pass
        base = self.base.base
        base = FP.base_plot(base)

    def plot_median(self):
        """
        Wrapper to plot medians of all lines

        """
        self.data.plot_median()

    # def zooming(self):
    #     pass

    # def zoomOut(self):
    #     pass

    # def zoomIn(self):
    #     pass

    # def zoomInitial(self):
    #     pass

    def save_plot(self):
        """
        Function saves plot inside actual window into a png file
        The file name is prefix_date_time.png
        Prefix depends on the actual image type.

        Returns
        -------
        None.

        """
        now = datetime.now()
        c_time = now.strftime("%H-%M-%S")
        today = date.today()
        d1 = today.strftime("%Y-%m-%d")
        fname = f"{self.data.data['type']}_" + f"{d1}_{c_time}.png"
        self.fig.savefig(fname)

    def diurnal(self):
        """
        If no base station data have been read create diurnal variation
        by adjusting a spline of degree deg to the median values as function
        of time.

        """
        if len(self.base_files) > 0:
            deg = 3
            self.base_flag = True
        else:
            results, okButton = dialog(
                ["No base station data exist.\nFit polygone to medians or "
                 + "cancel", "Degree of polynom"], ["l", "e"], [None, 5],
                "Diurnal variation by curve fit")
            if not okButton:
                print("\nDiurnal correction not applied")
            deg = int(results[1])
            self.base = DataContainer(0)
            self.base.base_ini()
            self.base_flag = False
        result = u.diurnal_correction(
            self.data, self.base.base, base_flag=self.base_flag, degree=deg)
        if result:
            self.treatments["diurnal"] = True
            self.data.set_values(treatments=self.treatments)
        self.w.basePlot.setEnabled(True)
        self.plotActual()

    def clean(self):
        """
        Wrapper to delete erroneous data

        Returns
        -------
        None.

        """
        self.data.clean()
        self.treatments["clip"] = True
        self.plotActual()

    def justify_median(self):
        """
        Wrapper to reduce directional effect by median adjustment

        """
        self.data.justify_median()
        self.treatments["justify_median"] = True
        self.plotActual()

    def justify_gauss(self):
        """
        Wrapper to reduce directional effect by Gaussian adjustment
        see Masoudi et al., J. Geophys. Eng., 2023

        Returns
        -------
        None.

        """
        self.data.justify_gauss()
        self.treatments["justify_Gauss"] = True
        self.plotActual()

    def interpol(self):
        """
        Interpolate data within the measurement lines onto a regular grid.
        No data are extrapolated, i.e. if a line starts later or finishes
        earlier than a regular grid, missing grid points are set to nan

        Returns
        -------
        None.

        """
        self.data.interpol()
        self.dat[self.actual_plotted_file] = deepcopy(self.data)
        self.treatments["interpol"] = True
        if "m" in self.data.data_type:
            self.w.poleReduction.setEnabled(True)
        self.w.tiltAngle.setEnabled(True)
        self.w.lineFFT.setEnabled(True)
        self.w.Spector_2D.setEnabled(True)
        self.w.prolongation.setEnabled(True)
        self.w.analytic.setEnabled(True)
        self.w.gaussJustify.setEnabled(True)
        self.w.originalData.setText("Restart with ungridded data")
        self.plotActual()

    def nan_fill(self):
        """
        Fill nan values by interpolation of data in the direction perpendicular
        to the measurement direction (it is supposed that if a line is not
        complete, nearby ones will be). Extrapolation will be done (one wants
        to create a complete grid) and different possibilities exist (mainly
        spline or constant). Spline is often very risky.

        Returns
        -------
        None.

        """
        self.data.nan_fill()
        self.plotActual()

    def reduce_pole(self):
        """
        Wrapper for pole reduction is done.

        Returns
        -------
        None.

        """
        self.data.reduce_pole()
        self.plotActual()

    def spector(self):
        """
        Wrapper for calculation of 2D Spector&Grant depths for all lines

        """
        self.data.spector()

    def spector2D(self):
        """
        Calculate depth of random sources with formula of (Spector and Grant,
        Geophysics, 1970) in 2D.

        Choose first a window length. Spectral coefficients are averaged over
        radial equidistant coefficients. Depths are calculated by fitting two
        lines to logarithmic spectrum. The break point between the two lines is
        searched between the 4th and the 10th spectral coefficient.
        Results of all lines are saved in file spector.dat.

        The window length should be defined such that the number of Fourier
        coefficients is at least 8:
        (n = window_length/(2*max(dx,dy)), dx, dy, step sizes defined during
        interpolation)

        Results are stored in file spector2D_<data_type>.dat; Data type may be
        "magnetic" or "gravity".

        """
        self.data.spector2D()

    def tilt(self):
        """
        Calculate tilt angle (Miller & Singh, JAG, 1994)

        Returns
        -------
        None.

        """
        self.lineaments = self.data.tilt()
        self.nlineaments = len(self.lineaments)
        if self.nlineaments:
            self.w.plotLineaments.setEnabled(True)
            self.w.plotLin_flag = True
            self.plot_lineaments()
        else:
            self.w.plotLineaments.setEnabled(False)
            self.w.plotLin_flag = False
        for d in self.dat:
            d.set_values(plotLin_flag=self.w.plotLin_flag)
        self.data.set_values(plotLin_flag=self.w.plotLin_flag)
        self.save_lineaments()
        self.plotActual()

    def continuation(self):
        """
        Wrapper to calculate field at higher or lower altitude
        """
        self.dz = self.data.continuation()
        self.treatments["up"] = True
        self.height += self.dz
        self.plotActual()

    def analytic_signal(self):
        """
        Calculate analytic signal (Nabighian, Geophysics, 1972)

        Returns
        -------
        None.

        """
        self.data.analytic_signal()

    def inver2D(self):
        """
        Do a 2D inversion along an interactively chosen line

        Returns
        -------
        None.

        """
        earth = deepcopy(self.earth)
        earth.dec -= self.data.data["line_declination"]
        data = []
        x = []
        y = []
        z = []
        data_type = self.data_types[self.actual_plotted_file][0]
        xx, yy, s1, s2, direction = self.plotLine(plot_flag=False)
        zz = np.ones_like(xx) * self.data.data["height"]
        if direction in ("N", "S", 0.0, 180.0):
            file = f"Inver2D_line_{int(yy)}m-E"
        else:
            file = f"Inver2D_line_{int(yy)}m-N"
            earth.dec -= 90.
        data.append(s1)
        x.append(xx)
        y.append(None)
        z.append(zz)
        if s2[0] is not None:
            x.append(xx)
            y.append(None)
            z.append(np.ones_like(xx) * self.data.data["height2"])
            data.append(s2)
        if self.data.topo_flag:
            inv = inversion(data, x, y, z, topo_int=self.data.topo_interpol,
                            earth=earth, data_type=data_type, line_pos=yy,
                            direction=direction)
        else:
            inv = inversion(data, x, y, z, earth=earth, data_type=data_type,
                            line_pos=yy, direction=direction)
# Get area of initial prisms and extract data to be inverted
        ret = inv.get_area2D()
        if not ret:
            return
# Define inversion parameters
        ret = inv.get_inversion_parameters(data_type)
        if not ret:
            return
# Set overall data and parameter variances
        ret = inv.get_variances()
        if not ret:
            return
        while True:
            inv.run_inversion()
            ret = inv.show_results2D(file)
            if not ret:
                break
            ret = inv.get_inversion_parameters(data_type)
            if not ret:
                break
            ret = inv.get_variances()
            if not ret:
                break
        inv.save_model()

    def inver3D(self):
        """
        Do a 3D inversion of data shown on screen

        Data must be gridded.

        It is possible to reduce the area to be inverted and the number of
        data points, taking one point out of n.

        The user will be asked interactively for the area to be inverted, for
        general inversion control parameters like the maximum number of
        iterations, other stopping conditions and smoothing and data and
        parameter variances.

        Returns
        -------
        None.

        """
        earth = deepcopy(self.earth)
        earth.dec -= self.data.data["line_declination"]
        if self.data.inter_flag:
            data_type = self.data_types[self.actual_plotted_file][0]
            data = []
            x = []
            y = []
            z = []
            data.append(self.data.sensor1_inter)
            xx, yy = np.meshgrid(self.data.x_inter, self.data.y_inter)
            x.append(xx)
            y.append(yy)
            z.append(np.ones_like(y) * self.data.data["height"])
            if self.data.data["grad_data"]:
                data.append(self.data.sensor2_inter)
                x.append(xx)
                y.append(yy)
                z.append(np.ones_like(yy) * (self.data.data["height2"]))
            inv = inversion(data, x, y, z, topo_int=None, earth=earth,
                            data_type=data_type)
# Define inversion parameters
            ret = inv.get_inversion_parameters(data_type)
            if not ret:
                return
# Set overall data and parameter variances
            ret = inv.get_variances()
            if not ret:
                return
# Get area of initial prisms and extract data to be inverted
            ret = inv.get_area3D()
            if not ret:
                return
# Do inversion
            while True:
                inv.run_inversion()
                ret = inv.show_results3D()
                if not ret:
                    break
                ret = inv.get_inversion_parameters(data_type)
                if not ret:
                    break
                ret = inv.get_variances()
                if not ret:
                    break
            inv.save_model()
            # os.remove("prism_control.dat")
            # os.remove("par_hist.dat")

        else:
            _ = QtWidgets.QMessageBox.warning(
                None, "Warning",
                "For 3D inversion, data must be interpolated onto a "
                + "regular grid.\n\nUse Utilities->Interpolate\n"
                + "and start inversion again",
                QtWidgets.QMessageBox.Close, QtWidgets.QMessageBox.Close)
            return

    def synthetic_model(self):
        """
        Define a 3D model composed of rectangular vertical prisms and calculate
        its effect. By default, the same type of data is calculated as the one
        of the read data (magnetic vs gravity and if magnetic, the same
        parameters for the Earth's field)')

        Returns
        -------
        None.

        """
        data_type = self.data_types[self.actual_plotted_file][0]
        x, y, z, sus, rem, rem_i, rem_d, rho = io.read_synthetic_model()
        if x is None:
            nprism = 0
            x1 = []
            x2 = []
            y1 = []
            y2 = []
            z1 = []
            z2 = []
            sus = []
            rem = []
            rem_i = []
            rem_d = []
            rho = []
            if "m" in data_type:
                s = 0.001
                r = 0.0
            else:
                s = 0.0
                r = 100.0
            while True:
                results, okButton = dialog(
                    [f"Prism {nprism+1}\n\n", "xmin [m]", "xmax [m]",
                     "ymin [m]", "ymax [m]", "zmin [m]", "zmax [m]",
                        "_________________________________",
                        "susceptibility [SI]", "remanence intensity [A/m]",
                        "remanence declination [°]",
                        "remanence inclination [°]", "density [kg/m3]"],
                    ["l", "e", "e", "e", "e", "e", "e", "l", "e", "e", "e",
                     "e", "e"],
                    ["b", 0.0, 50.0, 0.0, 50.0, 1.0, 5.0, None, s, 0.0, 0.0,
                     0.0, r], "Synthetic model")
                if not okButton:
                    if nprism == 0:
                        print("No prism defined, synthetic model calculation "
                              + "aborted")
                        return
                    break
                nprism += 1
                x1.append(float(results[1]))
                x2.append(float(results[2]))
                y1.append(float(results[3]))
                y2.append(float(results[4]))
                z1.append(float(results[5]))
                z2.append(float(results[6]))
                sus.append(float(results[8]))
                rem.append(float(results[9]))
                rem_i.append(float(results[10]))
                rem_d.append(float(results[11]))
                rho.append(float(results[12]))
            x = np.zeros((len(x1), 2))
            x[:, 0] = np.array(x1)
            x[:, 1] = np.array(x2)
            y = np.zeros((len(x1), 2))
            y[:, 0] = np.array(y1)
            y[:, 1] = np.array(y2)
            z = np.zeros((len(x1), 2))
            z[:, 0] = np.array(z1)
            z[:, 1] = np.array(z2)
            sus = np.array(sus)
            rem = np.array(rem)
            rem_i = np.array(rem_i)
            rem_d = np.array(rem_d)
            rho = np.array(rho)
            with open("synthetic_model.txt", "w") as fo:
                for i, s in enumerate(sus):
                    fo.write(
                        f"{x[i, 0]} {x[i, 1]} {y[i, 0]} {y[i, 1]} "
                        + f"{z[i, 0]} {z[i, 1]} {s} {rem[i]} {rem_i[i]} "
                        + f"{rem_d[i]} {rho[i]}\n")
        nprism = len(sus)
        xmin = self.data.xmin
        xmax = self.data.xmax
        ymin = self.data.ymin
        ymax = self.data.ymax
        dx = np.round((xmax - xmin) / 50.0, 0)
        dy = np.round((ymax - ymin) / 50.0, 0)
        labels = ["xmin [m]", "xmax [m]", "dx [m]", "__________________",
                  "ymin [m]", "ymax [m]", "dy [m]"]
        types = ["e", "e", "e", "l", "e", "e", "e"]
        values = [xmin, xmax, dx, None, ymin, ymax, dy]
        if "m" in data_type:
            labels += ["__________________\nif height2==height1 or"
                       + "height2==None:\n   only 1 sensor",
                       "height sensor 1", "height sensor 2"]
            types += ["l", "e", "e"]
            values += [None, 0.0, 0.0]
        results, okButton = dialog(labels, types, values, "Calculation points")
        if not okButton:
            print("Calculation of synthetic model aborted")
            return
        xmin = float(results[0])
        xmax = float(results[1])
        dx = float(results[2])
        ymin = float(results[4])
        ymax = float(results[5])
        dy = float(results[6])
        if len(labels) > 7:
            height1 = float(results[8])
            if results[9].lower() == "none":
                height2 = height1
            else:
                height2 = float(results[9])
        earth = deepcopy(self.earth)
        earth.dec -= self.data.data["line_declination"]
        xx = np.arange(xmin, xmax + dx / 2.0, dx)
        yy = np.arange(ymin, ymax + dy / 2.0, dy)
        xd, yd = np.meshgrid(xx, yy)
        d_shape = xd.shape
        ndat1 = d_shape[0] * d_shape[1]
        xdat = []
        ydat = []
        zdat = []
        data = []
        xdat.append(xd)
        ydat.append(yd)
        zdat.append(np.ones_like(xd) * height1)
        data.append(np.zeros_like(xd))
        if height2 != height1:
            data.append(np.zeros_like(xd))
            xdat.append(xd)
            ydat.append(yd)
            zdat.append(np.ones_like(xd) * height2)
        self.sPrism = PP(earth)
        for i in range(nprism):
            self.sPrism.add_prism(
                x[i], y[i], z[i], sus[i], rem[i], rem_i[i], rem_d[i], rho[i])
        inv = inversion(data, xdat, ydat, zdat, earth=earth,
                        data_type=data_type)
        inv.mPrism = self.sPrism
        inv.params = np.zeros(nprism + 1)
        if "m" in data_type:
            if abs(sus).max() > 0.0:
                inv.sus_inv = True
                inv.params[:nprism] = sus
            if abs(rem).max() > 0.0:
                inv.rem_inv = True
                inv.params[:nprism] = rem
        else:
            inv.rho_inv = True
            inv.params[:nprism] = rho
            inv.params
        inv.max_iter = 0
        if height1 != height2:
            inv.x = np.concatenate((xdat[0].flatten(), xdat[1].flatten()))
            inv.y = np.concatenate((ydat[0].flatten(), ydat[1].flatten()))
            inv.z = np.concatenate((zdat[0].flatten(), zdat[1].flatten()))
            inv.data = np.zeros_like(inv.x)
        inv.run_inversion()
        dat = io.Data(0)
        if height1 == height2:
            dat.store_gxf("synthetic_data.gxf", inv.data_mod.reshape(d_shape),
                          xmin, ymin, dx, dy)
        else:
            data1 = inv.data_mod[:ndat1].reshape(d_shape)
            data2 = inv.data_mod[ndat1:].reshape(d_shape)
            dat.write_dat("synthetic_data.dat", data1, xx, yy, data2=data2)
            with open("synthetic_data.config", "w") as fo:
                fo.write(f"{data_type}\n")
                fo.write("synthetic model\n")
                fo.write(f'{self.dat[-1].data["line_declination"]}\n')
                fo.write(f"{height1}\n")
                fo.write(f"{height2}\n")
                fo.write("0\n")
                fo.write(f"{self.earth.f}\n")
                fo.write(f"{self.earth.inc}\n")
                dec = self.earth.dec
                fo.write(f'{dec+self.dat[-1].data["line_declination"]}\n')
        inv.show_synthetic()

    def Handler(self, signal_received, frame):
        """
        Handles CTRL-C key stroke

        Parameters
        ----------
        signal_received : CTRL-C keystroke
        frame : No idea

        Returns
        -------
        None.

        """
        self.close_app()

    def save_lineaments(self):
        """
        If lineaments were picked, save the information into file
        lineaments.dat

        The file has the format:
        #key (including the character #)
        x y

        Actually, key may be "magnetic tilt" or "gravity tilt", meaning
        lineaments having been traced on tilt angle maps of gravity or
        megnetic data.

        x y line is repeated for every point defining a lineament.
        File is finished with "#END"

        Returns
        -------
        None.

        """
        if self.nlineaments > 0:
            with open("lineaments.dat", "w", encoding="utf-8") as fo:
                for lin in self.lineaments.values():
                    fo.write(f"#{lin['type']}\n")
                    for i in range(len(lin["x"])):
                        fo.write(f"{lin['x'][i]:0.1f}  {lin['y'][i]:0.1f}\n")
                fo.write("#END\n")

    def close_app(self):
        """
        Finishes application:

        Stores picks into file pick.dat
        Stores information concerning possibly modified traces
        (muted or sign-inversed)
        Deletes unneeded folder if tomography was calculated
        Closes window
        """
        choice = QtWidgets.QMessageBox.question(
            None, "Confirm", "Are you sure?",
            QtWidgets.QMessageBox.Yes | QtWidgets.QMessageBox.No)
        if choice == QtWidgets.QMessageBox.Yes:
            print("\nApplication finished.\n\n"
                  + "Close console if you are working with Spyder")
            self.w.close()
            QtWidgets.QApplication.quit()
            return True
        return False



# -*- coding: utf-8 -*-
"""
Created on Mon Aug 12 13:34:18 2024

@author: Hermann
"""

from pathlib import Path
from runpy import run_path

pkg_dir = Path(__file__).resolve().parent


def execute_script():
    script_pth = pkg_dir / "pymagra_script.py"
    run_path(str(script_pth), run_name="__main__")



# -*- coding: utf-8 -*-
"""
Last modified on Jan 14, 2025

@author: Hermann Zeyen <hermann.zeyen@universite-paris-saclay.fr>
         University Paris-Saclay, France

Contains functions for data input/output

Contains methods:
    - get_files
    - read_lineaments
    - read_synthetic_model
    - read_geography_file
    - get_mag_field
    - read_geometrics
    - write_geometrics

"""

#        fill_nans (actually not used)
#            eliminate_nans (actually not used)
import sys
import os
import numpy as np
from PyQt5 import QtWidgets

# from ..plotting import plot as w
from .dialog import dialog
from ..data.geometrics import Geometrics
from .earth import Earth_mag


def get_files(dir0=None, ftype=None):
    """
    Ask for files with ending "stn".
    Several files may be chosen at once using as usual SHFT or CTRL.
    If a folder was chosen, it is automatically recognized and eliminated
    from the list of file names.

    Returns
    -------
    data_files: list str
        list of chosen files

    """
    valid_extensions = np.array([".STN", ".OUT", ".XYZ", ".GXF", ".DAT"])
    ftypes = ("GEOMETRICS", "MGWIN", "BRGM", "GXF", "GEOMETRICS")
    dtypes = ("magnetic", "gravity")
    #    sensor2_types = ["GEOMETRICS", "MGWIN"]
    try:
        os.chdir(dir0)
    except (FileNotFoundError, TypeError):
        dir0 = None
    files = list(
        QtWidgets.QFileDialog.getOpenFileNames(
            None, "Select data files", "",
            filter="stn/gxf/XYZ/dat (*.stn *.gxf *.XYZ *.dat) ;; all (*.*)"))
    if len(files) == 0:
        print("No file chosen, program finishes")
        sys.exit("No file chosen")
    elif len(files[0]) == 0:
        print(
            "\nNo file chosen, program finishes\n\n"
            + "You probably must close the Spyder console before restarting")
        sys.exit("No file chosen")
    # Sort chosen file names
    files[0].sort()
    # Check data formats
    # Set working folder as folder of the first selected file
    if ftype != "base":
        dir0 = os.path.dirname(files[0][0])
        os.chdir(dir0)
        print(f"Set working folder to: {dir0}")

    # Loop over file names and store valid file names into list data_files
    data_files = []
    file_types = []
    data_types = []
    file_conf = []
    for _, f in enumerate(files[0]):
        _, file_ext = os.path.splitext(f)
        file_ext = file_ext.upper()
        if os.path.isdir(f):
            continue
        if file_ext not in valid_extensions:
            continue
        file_types.append(ftypes[np.where(valid_extensions == file_ext)[0][0]])
        data_files.append(f)
        fconfig = os.path.basename(f)
        j = fconfig.rfind(".")
        # Check whether there is a configuration file for each data file
        if j > 0:
            fconfig = fconfig[:j] + ".config"
        else:
            fconfig += ".config"
        # If there is, read its content
        if os.path.isfile(fconfig):
            with open(fconfig, "r") as fc:
                data_types.append(fc.readline()[:-1])
            file_conf.append(True)
        else:
            file_conf.append(False)
            data_types.append("")
    if len(data_files) == 0:
        _ = QtWidgets.QMessageBox.critical(
            None, "Error", "No valid data files given\n\n"
            + f"Only {valid_extensions} allowed.\n\nProgram stops",
            QtWidgets.QMessageBox.Ok)
        raise Exception("File type error.\n")
    # Ask for data types
    labels = []
    values = []
    types = []
    if ftype == "base":
        for _ in range(len(files[0])):
            data_types.append("magnetic")
    else:
        for i, f in enumerate(data_files):
            if file_conf[i]:
                continue
            labels.append(f"{os.path.basename(f)}:")
            values.append(None)
            types.append("l")
            labels.append(["Magnetic", "Gravity"])
            if "gra" in f.lower() or "bou" in f.lower():
                values.append("2")
            else:
                values.append("1")
            types.append("r")
        if len(labels) > 0:
            results, ok_button = dialog(labels, types, values,
                                        title="data types")
            if not ok_button:
                print("No entry, program finished")
                sys.exit()
            ir = -1
            for i, flag in enumerate(file_conf):
                if flag:
                    continue
                ir += 2
                data_types[i] = dtypes[int(results[ir])]
    return data_files, file_types, data_types, dir0


def read_lineaments(file="lineaments.dat"):
    """
    Reads file with picked lineament information (usually done on tilt angle
    plots)

    Parameters
    ----------
    file : str, optional. Default: "lineaments.dat"
        Name of file to be read

    Returns
    -------
    lineaments : dictionary
        The following entries are available:

        - "x" : numpy 1D float array
          Contains x coordinates of lineaments (E-W direction)
        - "y" : numpy 1D float array
          Contains y coordinates of lineaments (N-S direction)
        - "type" : str
          Type of lineament. May be "magnetic" or "gravity"

    """
    with open(file, "r", encoding="utf-8") as fi:
        lines = fi.readlines()
    il = 0
    nlineaments = 0
    x = []
    y = []
    lineaments = {}
    while True:
        if lines[il][0] == "#":
            if len(x) > 0:
                lineaments[nlineaments]["x"] = np.array(x)
                lineaments[nlineaments]["y"] = np.array(y)
                x = []
                y = []
            if lines[il][:4] == "#END":
                break
            nlineaments += 1
            lineaments[nlineaments] = {}
            lineaments[nlineaments]["type"] = lines[il][1:-1]
            il += 1
        else:
            nums = lines[il].split()
            x.append(float(nums[0]))
            y.append(float(nums[1]))
            il += 1
    return lineaments


def read_synthetic_model():
    """
    Read synthetic model.
    The file should have an extension .txt, .dat or .mod
    The model is composed of rectangular prisms with faces parallel to axis.
    The format of the file is as follows:

    - No header line
    - One line per prism to be calculated containing 7 to 11 values each:
      xmin, xmax, ymin, ymax, zmin, zmax, sus, rem_s, rem_i, rem_d, rho

    - xmin, xmax: minimum and maximum x_coordinates (E-W) of prism [m]
    - ymin, ymax: minimum and maximum y_coordinates (N-S) of prism [m]
    - zmin, zmax: minimum and maximum z_coordinates (positive down) of
      prism [m]
    - sus: susceptibility [SI units]
    - rem_s: intensity of remanent magnetization [A/m]
    - rem_i: inclination of remanent magnetization [degrees]
    - rem_d: declination of remanent magnetization [degrees]
    - rho: density of prism [kg/m3]

    Returns
    -------
    x : numpy float array of shape (n_prisms, 2)
        X-coordinates of prisms.
    y : numpy float array of shape (n_prisms, 2)
        Y-coordinates of prisms.
    z : numpy float array of shape (n_prisms, 2)
        Z-coordinates of prisms.
    sus : numpy float array of shape (n_prisms)
        Susceptibilities of prisms.
    rem : numpy float array of shape (n_prisms)
        Remanence intensities of prisms.
    rem_i : numpy float array of shape (n_prisms)
        Remanence inclinations of prisms.
    rem : numpy float array of shape (n_prisms)
        Remanence declinations of prisms.
    rho : numpy float array of shape (n_prisms)
        Densities of prisms.

    """
    file = list(
        QtWidgets.QFileDialog.getOpenFileName(
            None, "Select model file", "",
            filter="txt/dat/mod (*.txt *.dat *.mod) ;; all (*.*)"))
    if len(file) == 0:
        print("No file chosen, program finishes")
        return None, None, None, None, None, None, None, None
    if len(file[0]) < 1:
        print("read_synthetic_model: No files read")
        return None, None, None, None, None, None, None, None
    xmin = []
    xmax = []
    ymin = []
    ymax = []
    zmin = []
    zmax = []
    sus = []
    rem = []
    rem_i = []
    rem_d = []
    rho = []
    with open(file[0], "r") as fi:
        lines = fi.readlines()
    for line in lines:
        val = line.split()
        ncol = len(val)
        if ncol < 7:
            answer = QtWidgets.QMessageBox.warning(
                None,
                "Warning",
                "Synthetic model file does not have enough columns:\n"
                + f"At least 7 columns are needed, {ncol} found.\n"
                + "Synthetic modeling aborted.",
                QtWidgets.QMessageBox.Close, QtWidgets.QMessageBox.Ignore)
            return None, None, None, None, None, None, None, None
        if ncol < 11:
            if ncol == 7:
                text = "Remanence and density are set to zero."
            else:
                text = "Density is set to zero."
            answer = QtWidgets.QMessageBox.warning(
                None,
                "Warning",
                "Synthetic model file has only {ncol} columns:\n"
                + f"{text}\nPress Ignore to accept or Abort to abandon.",
                QtWidgets.QMessageBox.Ignore | QtWidgets.QMessageBox.Abort,
                QtWidgets.QMessageBox.Ignore)
            if answer == QtWidgets.QMessageBox.Abort:
                return None, None, None, None, None, None, None, None
        xmin.append(float(val[0]))
        xmax.append(float(val[1]))
        ymin.append(float(val[2]))
        ymax.append(float(val[3]))
        zmin.append(float(val[4]))
        zmax.append(float(val[5]))
        sus.append(float(val[6]))
        if ncol > 7:
            rem.append(float(val[7]))
            if ncol > 8:
                rem_i.append(float(val[8]))
                if ncol > 9:
                    rem_d.append(float(val[9]))
                    if ncol > 10:
                        rho.append(float(val[10]))
                    else:
                        rho.append(0.0)
                else:
                    rem_d.append(0.0)
                    rho.append(0.0)
            else:
                rem_i.append(0.0)
                rem_d.append(0.0)
                rho.append(0.0)
        else:
            rem.append(0.0)
            rem_i.append(0.0)
            rem_d.append(0.0)
            rho.append(0.0)
        nprism = len(xmin)
        x = np.zeros((nprism, 2))
        x[:, 0] = np.array(xmin)
        x[:, 1] = np.array(xmax)
        y = np.zeros((nprism, 2))
        y[:, 0] = np.array(ymin)
        y[:, 1] = np.array(ymax)
        z = np.zeros((nprism, 2))
        z[:, 0] = np.array(zmin)
        z[:, 1] = np.array(zmax)
    return x, y, z, np.array(sus), np.array(rem), np.array(rem_i), \
        np.array(rem_d), np.array(rho)


def read_geography_file(file):
    """
    Reads file with geography information to be plotted
    (borders - mainly geological, towns)

    Parameters
    ----------
    file : str, name of file to be read
        File has the following structure:

        - keyword may be "#POINT", "#LINE" or "#END"
        - if keyword == "#POINT", one line follows with x, y coordinates and
          text, text being the description of the point (no blanks)
        - if keyword == "#LINE", several lines follow, each one with x and y
          coordinate of one point describing the line
        - if keyword == "#END", this line finishes the data entry, possible
          following lines will be ignored.

    Returns
    -------
    geography : dictionary with all geography information.
        key is consecutive numbering of entries
        Each entry consists of a dictionary with the following entries:

        - "type" str
          may be "POINT" or "LINE"

          - If type == "POINT" : One line with:

              - "x" : float: x coordinate of point (East)
              - "y" : float: y coordinate of point (North)
              - "name" : str: Text to be plotted beside the point mostly name
                of a town

          - If type == "line" :

              - "x" : list of floats, East coordinate of points describing
                the line
              - "y" : list of floats, North coordinate of points describing
                the line

    """
    with open(file, "r", encoding="utf-8") as fi:
        ll = fi.readlines()
    geography = {}
    il = 0
    iunit = -1
    while True:
        if ll[il].upper().startswith("#POINT"):
            iunit += 1
            il += 1
            nums = ll[il].split()
            geography[iunit] = {}
            geography[iunit]["type"] = "POINT"
            geography[iunit]["x"] = float(nums[0])
            geography[iunit]["y"] = float(nums[1])
            geography[iunit]["name"] = nums[2]
            il += 1
        elif ll[il].upper().startswith("#LINE"):
            iunit += 1
            geography[iunit] = {}
            geography[iunit]["type"] = "LINE"
            geography[iunit]["x"] = []
            geography[iunit]["y"] = []
            while True:
                il += 1
                if ll[il].startswith("#"):
                    break
                nums = ll[il].split()
                geography[iunit]["x"].append(float(nums[0]))
                geography[iunit]["y"].append(float(nums[1]))
        elif ll[il].upper().startswith("#END"):
            break
    return geography


def get_mag_field(line_dir, strength=None, inclination=None, declination=None):
    """
    Get parameters of Earth's magnetic field in the study area
    The declination is calculated as absolute declitation minus line direction

    Parameters
    ----------
    line_dir : float
        Direction of Y axis with respect to North (positive towards E, degree)
    strength: float, optional. Default: NOne
        Field strength [nT]. If None, the field parameters are asked for
        interactively
    inclination: float, optional. Default: NOne
        Field inclination [degrees].
    declination: float, optional. Default: NOne
        Field declination [degrees] with respect to geographic north.

    Returns
    -------

    inclination : float
        Inclination of Earth's field in degrees
    declination : float
        Declination of Earth's field in degrees
    strength : float
        Strength of magnetic field [nT]
    """
    if strength is None:
        results, _ = dialog(
            ["Field strength [nT]", "Field inclination", "Field declination"],
            ["e", "e", "e"], [50000, 62, 0], "Earth's field parameters")
        strength = float(results[0])
        inclination = float(results[1])
        declination = float(results[2]) - line_dir
    else:
        declination -= line_dir
    earth = Earth_mag(strength, inclination, declination)
    return earth


def read_geometrics(file, n_block, height1, height2, dispo):
    """
    Read Geometrics .stn or .dat file (G-858 instrument)

    Parameters
    ----------
    file : str
        Name of data file.
    n_block : int
        Number of block to be read
    height1 : float
        Height of sensor 1 above ground (meters)
    height2 : float
        Height of sensor 2 above ground (meters)
    dispo : int
        Disposition of sensors if there are two sensors:
        0: vertical disposition, 1: horizontal disposition

    Returns
    -------
    gdata :  Dictionary with keys equal to line numbers (starting at 0)
        Each line is itself a dictionary with the following entries:
        Key is line number

        - "s1": Numpy float array with data of sensor 1
        - "s2": Numpy float array with data of sensor 2
          If only data of one single sensor were measured, "s2" contains
          only one zero.
        - "x":  Numpy float array with E-W coordinates of data points
        - "y":  Numpy float array with N-S coordinates of data points
        - "grad_flag" bool. True if 2 sensors were used, False if only
          one sensor
        - "mask": bool, True if line should be plotted, False if excluded
          from plotting. Initially set to True

    The original data are stored in class geometrics.Geometrics. See file
    geometrics.py for documentation

    """
    gdata = Geometrics()
    if ".stn" in file:
        gdata.read_stn(file, n_block, height1, height2, dispo)
    else:
        gdata.read_dat(file, n_block, height1, height2, dispo)
        os.remove("temp.stn")
    return gdata


def write_geometrics(self, file, data1, x, y, data2=None, n_block=0,
                     time=None):
    """
    Wrapper to write data in Geometrics MagMap2000 .stn format.

    Data must be interpolated onto a regular grid.

    Parameters
    ----------
    file : str
        File name where to write data.
    data1 : numpy float array [number_of_samples_per_line, number_of_lines]
        Data of sensor 1.
    x : numpy float array [number_of_samples_per_line]
        X coordinates of all measured points.
    y : numpy float array [number_of_samples_per_line, number_of_lines]
        Y coordinates of all measured points.

    Optional parameters:

    data2 : numpy float array [number_of_samples_per_line, number_of_lines]
        Data of sensor 2. Optional. Default: np.zeros_like(data1)
    n_block : int, optional
        Number of block (data set) to be written. The default is 0.

    Returns
    -------
    None.

    """
    if not isinstance(data1, np.ndarray):
        n_block = data1["block"] - 1
    n_block = 0
    self.gdata[n_block].write_stn(file, data1, x, y, data2=data2, time=time)
